#include <vector.h>

#include <vector3.h>
#include <plane.h>

#include <string.h>
#include <assert.h>
#include <stdio.h>

//
//
//

class ConvexPolygon 
{
public:

	ConvexPolygon();

	// Construction

	void addPoint(const Vector &pnt);
	void constructionDone();

	//

	int empty() const;
	int inside(const Vector &pnt) const;

	//

	friend ConvexPolygon         intersection(const ConvexPolygon &a,const ConvexPolygon &b);
	friend vector<ConvexPolygon> difference  (const ConvexPolygon &a,const ConvexPolygon &b);

	friend void clipConvexPolygon
	(
		const ConvexPolygon &poly,
		const Plane         &plane,
			  ConvexPolygon &inside,
			  ConvexPolygon &outside
	);

	//

	friend ostream &operator <<(ostream &os,const ConvexPolygon &polygon);

private:
	
	vector<Vector>   _point;
	vector<Plane>    _plane;

	int              _done;
};

ConvexPolygon::ConvexPolygon()
: _done(0)
{
	_point.reserve(3);
	_plane.reserve(3);
}

void
ConvexPolygon::addPoint(const Vector &pnt)
{
	assert(!_done);
	_point.push_back(pnt);
}

void
ConvexPolygon::constructionDone()
{
	assert(!_done);
	_done = 1;

	assert(_point.size()>2);

	for (int i=0; i<_point.size(); i++)
	{
		int j = (i+1)%_point.size();

		// Build plane

		Vector normal = xProduct(_point[j] - _point[i],VectorZ);
		Plane  plane(_point[i],normal);

		// Flip the plane normal if necessary

		for (int k=0; k<_point.size(); k++)
		{
			if (k==i || k==j) continue;

			if (plane.inside(_point[k]))
				normal = -normal;
		}
		
		//

		_plane.push_back(Plane(_point[i],normal));
	}
}

int 
ConvexPolygon::empty() const
{
	return !_point.size();
}

int 
ConvexPolygon::inside(const Vector &pnt) const
{
	assert(_done);

	for (int i=0; i<_plane.size(); i++)
		if (_plane[i].inside(pnt))
			return 0;
		
	return 1;
}

ostream &
operator <<(ostream &os,const ConvexPolygon &polygon)
{
	os << "ConvexPolygon of " << polygon._point.size();
	os << " points and " << polygon._plane.size() << " planes." << endl;

	for (int p=0; p<polygon._point.size(); p++)
		os << polygon._point[p] << endl;

	return os;
}

//
//

ConvexPolygon 
intersection(const ConvexPolygon &a,const ConvexPolygon &b)
{
	ConvexPolygon result = a;

	for (int p=0; p<b._plane.size(); p++)
	{
		ConvexPolygon clipped;
		clipConvexPolygon(result,b._plane[p],ConvexPolygon(),clipped);
		result = clipped;
	}

	return result;
}

vector<ConvexPolygon> 
difference(const ConvexPolygon &a,const ConvexPolygon &b)
{
	vector<ConvexPolygon> result;

	for (int p=0; p<b._plane.size(); p++)
	{
		ConvexPolygon clipped;
		clipConvexPolygon(a,b._plane[p],clipped,ConvexPolygon());

		if (!clipped.empty())
			result.push_back(clipped);
	}

	return result;
}

//
//
//

void 
clipConvexPolygon
(
	const ConvexPolygon &poly,
	const Plane         &plane,
	      ConvexPolygon &inside,
		  ConvexPolygon &outside
)
{
	const int points = poly._point.size();

	for (int i=0; i<points; i++)
	{
		int j = (i+1)%points;

		const Vector &p1 = poly._point[i];
		const Vector &p2 = poly._point[j];

		int ic = plane.inside(p1);
		int jc = plane.inside(p2);

		if (ic)
			inside.addPoint(p1);
		else
			outside.addPoint(p1);

		if (ic!=jc)
		{
			Vector intersection = plane.intersectRay(p1,p2);

			inside.addPoint(intersection); 
			outside.addPoint(intersection); 
		}
	}

	if (!inside.empty())
		inside.constructionDone();

	if (!outside.empty())
		outside.constructionDone();
}

const int colors = 7;
unsigned char color[colors][3] =
{
	{   0,   0,   0 },
	{ 255,   0,   0 },
	{   0, 255,   0 },
    {   0,   0, 255 },
	{ 255, 255,   0 },
	{ 255,   0, 255 },
	{   0, 255, 255 }
};

int main(int argc,char *argv[])
{
	//

	ConvexPolygon polygon1;

	polygon1.addPoint(Vector(0.1,0.1,0.0));
	polygon1.addPoint(Vector(0.9,0.2,0.0));
	polygon1.addPoint(Vector(0.7,0.9,0.0));
	polygon1.addPoint(Vector(0.1,0.8,0.0));

	polygon1.constructionDone();

	//

	ConvexPolygon polygon2;

	polygon2.addPoint(Vector(0.5,0.2,0.0));
	polygon2.addPoint(Vector(0.9,0.8,0.0));
	polygon2.addPoint(Vector(0.2,0.6,0.0));

	polygon2.constructionDone();

	//

	ConvexPolygon polygon3 = intersection(polygon1,polygon2);

	//

	vector<ConvexPolygon> polygon4 = difference(polygon1,polygon2);
	vector<ConvexPolygon> polygon5 = difference(polygon2,polygon1);

	//

	cout << "POLYGON4" << endl;
	for (int p=0; p<polygon4.size(); p++)
		cout << polygon4[p];
	cout << endl;

	cout << "POLYGON5" << endl;
	for (p=0; p<polygon5.size(); p++)
		cout << polygon5[p];
	cout << endl;

	const int mapRes = 500;
	FILE *fout = fopen("out.ppm","wb");

	unsigned char buffer[1024];
	sprintf((char *) buffer,"P6 %d %d 255\n",mapRes,mapRes);

	fwrite(buffer,1,strlen((char *) buffer),fout);

	for (int y=mapRes-1; y>=0; y--) 
		for (int x=0; x<mapRes; x++) 
		{
			Vector pos = VectorX*0.002*x+VectorY*0.002*y; 

/*
			int c = (
				polygon5[0].inside(pos) ? 1 :
				polygon5[1].inside(pos) ? 2 : 0 );
*/
			int c = polygon3.inside(pos);
			fwrite(color[c],1,3,fout);
		}

	fclose(fout);
}
