//
// OpenGL CSG Rendering
//
// Nigel Stewart (nigels@eisa.net.au)
// RMIT Department of Mechanical and Manufacturing Engineering
// 27th December 1998
//

#include <csgtree.h>
#include <parsetr.h>

#include <timer.h>
#include <glcursor.h>
#include <font.h>
#include <colours.h>
#include <trees.h>

#include <glbuffer.h>	
#include <cntsurf.h>

#include <subtract.h>
#include <scs.h>
#include <goldfeat.h>

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include <iostream.h>
#include <fstream.h>

#ifdef WIN32
#include <strstrea.h>
#else
#include <strstream.h>
#endif

#ifdef WIN32
#pragma warning(disable : 4244)     // Data conversion 
#pragma warning(disable : 4305)     // Truncation  
#pragma warning(disable : 4018)     // Signed/unsigned 
#endif

//
//

extern char help[];
extern void CalcKmap(const Uint mapRes,GLuint dl,char *filename);
extern void OpenGLinfo(ostream &os);

void makeObjectMenu();

//
// Variables
//

typedef enum
{
	DRAW_PREVIEW,
	DRAW_DEPTH_RANGE,
	DRAW_WIREFRAME,
	DRAW_CONVEXITY,
	DRAW_GOLDFEATHER,
	DRAW_GOLDFEATHER_LAYERS,
	DRAW_SCS
} RenderingMode;

RenderingMode display = DRAW_PREVIEW;

// Misc

bool capture = false;
bool kMap = false;

// Viewing direction and animation clock

Matrix   viewing;

Timer    worldTime;

typedef enum
{
	ANIMATE_MODE_NONE = 0,			// Disable Animation
	ANIMATE_MODE_WALL_CLOCK,		// Position determined by real time
	ANIMATE_MODE_FRAME				// Frame-rate invariant
} AnimationType;

AnimationType animate = ANIMATE_MODE_WALL_CLOCK;

// CSG Configuration

bool  copyPixels = false;

// Viewport

GLint size = 200;

// On-screen info

bool  showInfo = 0;
bool  showFrameRate = 0;

// Menus

int treeSubMenu = 0;
int objectSubMenu = 0;
int toolsSubMenu =0;
int configSubMenu = 0;
int benchmarkSubMenu = 0;
int animateSubMenu = 0;
int mainMenu = 0;

// BenchMarking

Timer benchmarkTimer;
Uint  benchmarkInProgress = 0;
Uint  framesRemaining = 0;

////////////

Uint frameCounter = 0;
extern void	setConvexTable();

////////////


//
//
//

vector<CsgTreeNode *> tree;
int                   currentTree = 0;
CsgPrimitiveDatabase  *database = NULL;

void
SelectTree(const int t)
{
	if ( t<tree.size() )
	{
		currentTree = t;
		tree[currentTree]->normalise();
	}
}

bool
LoadTree(istream &is)
{
	if (!is.good())
		return false;

	if (database)
		delete database;
	database = new CsgPrimitiveDatabase();

	vector<CsgTreeNode *> tmp = parseTrees(*database,is);

	if (tmp.size()>0)
	{
		tree = tmp;
		SelectTree(0);
		makeObjectMenu();

		return true;
	}

	cout << "Error Parsing CSG Tree." << endl;
	return false;
}

bool
LoadTree(char *tree)
{
	istrstream fin(tree);
	return LoadTree(fin);
}

//
//
//

void 
Init()
{
}

void 
Draw()
{
	//
	// Calculate frame rate, smoothing over last four frames
	//

	static Timer  frameTimer;
	static double recentTimes[4] = { 0.0, 0.0, 0.0, 0.0 };
	static int    currentTime = 0;

	recentTimes[currentTime] = frameTimer.elapsed();
	currentTime = (currentTime+1)%4;

	int n = 4;
	for (int t=0;t<4;t++)
		if (recentTimes[t]==0.0)
			n--;

	double elapsedTime = max(recentTimes[0],max(recentTimes[1],max(recentTimes[2],max(0.01,recentTimes[3]))));
	frameTimer.start();

	if (tree.size())
	{
		if (display==DRAW_PREVIEW)
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
		else
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
//		gluPerspective(45, 1.33, 0.8, 10.0);

		glScalef(1,1,-0.2);

		glMatrixMode(GL_MODELVIEW);		
		glLoadIdentity();

		//

		glEnable(GL_NORMALIZE);
		glEnable(GL_DEPTH_TEST);
		glDisable(GL_DITHER);

		//
		// Viewing Settings
		//

		static GLfloat lmodel_ambient[] =     { 0.0, 0.0, 0.0, 0.0 };
		static GLfloat lmodel_twoside[] =     { GL_TRUE };
		static GLfloat lmodel_local[] =       { GL_FALSE};
		static GLfloat light0_ambient[] =     { 0.1, 0.1, 0.1, 1.0};
		static GLfloat light0_diffuse[] =     { 1.0, 1.0, 1.0, 0.0};
		static GLfloat light0_position[] =    { 0.2, 0.5, 1, 0};
		static GLfloat light0_specular[] =    { 1.0, 1.0, 1.0, 0.0};
		static GLfloat bevel_mat_ambient[] =  { 0.0, 0.0, 0.0, 1.0};
		static GLfloat bevel_mat_shininess[] = { 40.0};
		static GLfloat bevel_mat_specular[] = {1.0, 1.0, 1.0, 0.0};
		static GLfloat bevel_mat_diffuse[] = {1.0, 0.0, 0.0, 0.0};

		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);
		glEnable(GL_DEPTH_TEST);
		glClearDepth(1.0);

		glClearColor(0.5, 0.5, 0.5, 0.0);
		
		glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient);
		glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
		glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);
		glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
		glEnable(GL_LIGHT0);

		glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, lmodel_local);
		glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lmodel_twoside);
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
		glEnable(GL_LIGHTING);

		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, bevel_mat_ambient);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, bevel_mat_shininess);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, bevel_mat_specular);
		glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, bevel_mat_diffuse);

		glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);
		glShadeModel(GL_SMOOTH);

//		gluLookAt(0,0,5, 0,0,0, 0,1,0);
//		glTranslatef(0,0,5);

		viewing.glMultMatrix();

		switch (animate)
		{
		case ANIMATE_MODE_NONE:
			break;
		case ANIMATE_MODE_WALL_CLOCK:
			glRotated(worldTime.elapsed()*360.0/5.0,0.0,1.0,0.0);
			glRotated(worldTime.elapsed()*360.0/25.0,1.0,0.0,0.0);
			break;
		case ANIMATE_MODE_FRAME:
			{
				frameCounter = (frameCounter+1)%1000;
				glRotated(double(frameCounter)*360.0/200.0,0.0,1.0,0.0);
				glRotated(double(frameCounter)*360.0/1000.0,1.0,0.0,0.0);
			}
			break;
		}

//		glScaled(zoom,zoom,zoom);

		switch (display)
		{
		case DRAW_PREVIEW:
			tree[currentTree]->draw();
			break;

		case DRAW_DEPTH_RANGE:
			tree[currentTree]->draw();
			extern void zBufferVis();
			zBufferVis();
			break;

		case DRAW_WIREFRAME:
			glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
			glEnable(GL_CULL_FACE);
			glCullFace(GL_BACK);
			tree[currentTree]->draw();
			glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
			break;

		case DRAW_GOLDFEATHER:
			GoldfeatherRenderingAlgorithm(tree[currentTree]->sumOfProducts(),copyPixels);
			break;

		case DRAW_GOLDFEATHER_LAYERS:
			LayeredGoldfeatherRenderingAlgorithm(tree[currentTree]->sumOfProducts(),copyPixels);
			break;

		case DRAW_SCS:
			SequencedConvexSubtraction(tree[currentTree]->sumOfProducts(),copyPixels);			
			break;

		case DRAW_CONVEXITY:
			{
				GLuint everyThing = glGenLists(1);
				glNewList(everyThing,GL_COMPILE);
					tree[currentTree]->draw();
				glEndList();

				countSurfacesVisualise(&everyThing,1);

				glDeleteLists(everyThing,1);
			}
			break;
		}
//

//

		if (framesRemaining)
		{
			if (--framesRemaining==0)
			{
				cout << benchmarkInProgress << " frames in " << benchmarkTimer.elapsed() << " sec." << endl; 
				benchmarkInProgress = 0;
			}
		}

// 
		if (capture)
		{
			static char *colourFileName =  "0color.ppm";
			static char *depthFileName =   "0depth.ppm";
			static char *stencilFileName = "0stencil.ppm";

			{
				OpenGLcolourBuffer c;
				c.writePPM(colourFileName);
			}
			{
				depthBuffer d; 
				// OpenGLdepthBuffer<GLuint, GL_UNSIGNED_INT> d;
				d.writePPM(depthFileName);
			}

			(*colourFileName)++;
			(*depthFileName)++;
			(*stencilFileName)++;

			if (animate==ANIMATE_MODE_NONE || *colourFileName>'9')
				capture = 0;

			if (!capture)
				*colourFileName = *depthFileName = *stencilFileName = '0';
		}

		// Draw text info

		glPushAttrib(GL_ENABLE_BIT);

			glDisable(GL_LIGHTING);

			glCursorInit();
			glCursorHome();

			char buffer[1024];

			if (showFrameRate)
			{
				sprintf(buffer,"%2.1lf",1.0/elapsedTime);
				glColor3fv(colourWhite); 
				glCursorPrintNl(buffer);
				glCursorPrintNl(NULL);
			}

			if (showInfo)
			{
				ostrstream os(buffer,1024);
				os << "CSG Expression: " << *tree[currentTree] << '\0';
				glColor3fv(colourWhite);

				glCursorPrintNl(buffer);
	
				{
					char buffer[1024];
					ostrstream os(buffer,1024);
					os << "Normalised    : " << tree[currentTree]->sumOfProducts() << '\0';
	
					glCursorPrintNl(buffer);
				}

//				if (*useCopy)
//					glCursorPrintNl("(Copy depth buffers.)\n");
//				else
//					glCursorPrintNl("(Draw/Read depth buffers.)\n");
			}

		glPopAttrib();

//
//
		glFlush();
		glutSwapBuffers();
		glutPostRedisplay();
	}
}

void Reshape(int wx,int wy)
{
	size = wy;
	glViewport(0,0,wx,wy);
	glutPostRedisplay();

	cout << "Resize Window " << wx << 'x' << wy << endl;
}

void Key(unsigned char key,int x,int y)
{
	int index = key - '1';

	if (index>=0 && index<tree.size() && tree[index] && key<=int('9'))
		SelectTree(index);
	else
		switch (key)
		{
		case ' ':  display = (display==DRAW_PREVIEW ? DRAW_SCS : DRAW_PREVIEW); break;

		case 'h':  cout << help << endl; break;
		case 's':  animate = ( animate==ANIMATE_MODE_NONE ? ANIMATE_MODE_WALL_CLOCK : ANIMATE_MODE_NONE);   break;
		case 'q':  exit(0);              break;

		case 'i':  viewing = viewing * matrixScale(1.2);     break;
		case 'o':  viewing = viewing * matrixScale(1.0/1.2); break;
		case 'a':  viewing = viewing * matrixRotate(VectorX,10.0);  break;
		case 'z':  viewing = viewing * matrixRotate(VectorX,-10.0); break;
		case 'n':  viewing = viewing * matrixRotate(VectorY,10.0);  break;
		case 'm':  viewing = viewing * matrixRotate(VectorY,-10.0); break;
		
		case 'd':  viewing = viewing * matrixTranslate(-VectorZ*0.1); break;
		case 'c':  viewing = viewing * matrixTranslate( VectorZ*0.1); break;
		}
}

#define MENU_QUIT					1000

#define MENU_TOGGLE_INFO			1002
#define MENU_TOGGLE_FRAMERATE		1003
#define MENU_TOGGLE_COPY			1004
#define MENU_TOGGLE_CAPTURE			1005

#define MENU_SELECT_DEFAULT_TREE       1010
#define MENU_SELECT_GRID_TREE          1011
#define MENU_SELECT_ROW_TREE           1012
#define MENU_SELECT_HOLLOW_TREE        1013
#define MENU_SELECT_HELIX_TREE         1014
#define MENU_SELECT_SEG_HELIX_TREE     1015
#define MENU_SELECT_SWISS_CHEESE_TREE  1016
#define MENU_SELECT_PIPE_TREE          1017

#define MENU_DRAW_PREVIEW	        1020
#define MENU_DRAW_COUNT_SURFACES	1021
#define MENU_DRAW_GOLDFEATHER	    1022
#define MENU_DRAW_GOLDFEATHER_LAYER	1023
#define MENU_DRAW_SCS	            1024
#define MENU_DRAW_WIREFRAME         1025
#define MENU_DRAW_DEPTH_RANGE		1026
#define MENU_DRAW_KMAP              1029

#define MENU_BENCHMARK_10			1030
#define MENU_BENCHMARK_100			1031
#define MENU_BENCHMARK_1000			1032

#define MENU_ANIMATE_MODE_NONE			1040
#define MENU_ANIMATE_MODE_WALL_CLOCK	1041
#define MENU_ANIMATE_MODE_FRAME			1042

#define MENU_WRITE_POVRAY_OUTPUT    1050

#define MENU_TREE					0

void
menu(int value)
{
	int treeValue = value-MENU_TREE;

	if (treeValue>=0 && treeValue<tree.size() && tree[treeValue])
		SelectTree(treeValue);
	else
		switch (value)
		{
		case MENU_QUIT:
			exit(0);
			break;
		case MENU_TOGGLE_INFO:           showInfo = !showInfo;   break;
		case MENU_TOGGLE_FRAMERATE:	     showFrameRate = !showFrameRate; break;
		case MENU_TOGGLE_CAPTURE:        capture = !capture; break;

		case MENU_TOGGLE_COPY:			 
			copyPixels = !copyPixels; 
			if (copyPixels)
				glutReshapeWindow(size*2,size);
			else
				glutReshapeWindow(size,size);
			break;

		case MENU_DRAW_PREVIEW:           display = DRAW_PREVIEW; break;
		case MENU_DRAW_WIREFRAME:         display = DRAW_WIREFRAME; break;
		case MENU_DRAW_COUNT_SURFACES:    display = DRAW_CONVEXITY; break;
		case MENU_DRAW_GOLDFEATHER:       display = DRAW_GOLDFEATHER; break;
		case MENU_DRAW_GOLDFEATHER_LAYER: display = DRAW_GOLDFEATHER_LAYERS; break;
		case MENU_DRAW_SCS:               display = DRAW_SCS; break;
		case MENU_DRAW_DEPTH_RANGE:       display = DRAW_DEPTH_RANGE; break;

		case MENU_DRAW_KMAP:         CalcKmap(100,0,"kmap.ppm"); break;

		case MENU_SELECT_DEFAULT_TREE:   LoadTree(defaultTree); break;
		case MENU_SELECT_GRID_TREE:      LoadTree(gridTree); break;
		case MENU_SELECT_ROW_TREE:       LoadTree(rowTree); break;
		case MENU_SELECT_HOLLOW_TREE:    LoadTree(hollowTree); break;
		case MENU_SELECT_HELIX_TREE:     LoadTree(helixTree); break;
		case MENU_SELECT_SEG_HELIX_TREE: LoadTree(segmentedHelixTree); break;
		case MENU_SELECT_SWISS_CHEESE_TREE: LoadTree(swissCheeseTree); break;
		case MENU_SELECT_PIPE_TREE:         LoadTree(pipeTree); break;

		case MENU_BENCHMARK_10:
			if (!benchmarkInProgress)
			{
				benchmarkTimer.start();
				benchmarkInProgress = framesRemaining = 10;
			}
			break;

		case MENU_BENCHMARK_100:
			if (!benchmarkInProgress)
			{
				benchmarkTimer.start();
				benchmarkInProgress = framesRemaining = 100;
			}
			break;

		case MENU_BENCHMARK_1000:
			if (!benchmarkInProgress)
			{
				benchmarkTimer.start();
				benchmarkInProgress = framesRemaining = 1000;
			}
			break;

		case MENU_ANIMATE_MODE_NONE:		animate = ANIMATE_MODE_NONE; break;
		case MENU_ANIMATE_MODE_WALL_CLOCK:  animate = ANIMATE_MODE_WALL_CLOCK; break;
		case MENU_ANIMATE_MODE_FRAME:       animate = ANIMATE_MODE_FRAME; break;

		case MENU_WRITE_POVRAY_OUTPUT:
			if (database && tree[currentTree])
			{
				fstream fout("out.pov",ios::out);
				database->writePOV(fout,*(tree[currentTree]));
			}
			break;
		}
}

void makeMainMenu()
{
	mainMenu = glutCreateMenu(menu);
	glutAddSubMenu("Model",   treeSubMenu);
	glutAddSubMenu("Tree", objectSubMenu);
	glutAddSubMenu("Draw",  toolsSubMenu);
	glutAddSubMenu("Animate", animateSubMenu);
	glutAddSubMenu("Benchmark", benchmarkSubMenu);
	glutAddSubMenu("Config", configSubMenu);
	glutAddMenuEntry("Quit"       , MENU_QUIT);
	glutAttachMenu(GLUT_RIGHT_BUTTON);
}

void makeObjectMenu()
{
	objectSubMenu = glutCreateMenu(menu);
	const int maxLabelSize = 25;
    int primCount = 0;

	for (int t=0; t<tree.size(); t++)
	{
		char buffer[1024];
		ostrstream os(buffer,1024);
		os << *tree[t]<< '\0';

		if (strlen(buffer)>maxLabelSize)
			sprintf(buffer+maxLabelSize,"...");

		if (tree[t]->type()==PRIMITIVE_NODE)
		{
			primCount++;
			if (primCount>20)
				continue;
		}

		glutAddMenuEntry(buffer, MENU_TREE+t);
	}

	makeMainMenu();
}

////////////

int main(int argc,char *argv[])
{
	setConvexTable();

	cout << help << endl;

	viewing = viewing * matrixScale(0.8);

	glutInit(&argc,argv);
	glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL );

	glutInitWindowPosition(0,0);

	glutInitWindowSize(size,size);
	glutCreateWindow("OpenGL CSG Rendering");

	glutDisplayFunc(Draw);
	glutReshapeFunc(Reshape);

	fontInit();
	Init();

	OpenGLinfo(cout);
	cout << endl; 

	for (int a=1; a<argc; a++)
	{
		fstream fin(argv[a],ios::in | ios::nocreate);
		LoadTree(fin);
	}

	if (!tree.size())
		if (!LoadTree(defaultTree))
			return 0;

	glutKeyboardFunc(Key);

	treeSubMenu = glutCreateMenu(menu);
	glutAddMenuEntry("Default"    , MENU_SELECT_DEFAULT_TREE);
	glutAddMenuEntry("Row"        , MENU_SELECT_ROW_TREE);
	glutAddMenuEntry("Grid"       , MENU_SELECT_GRID_TREE);
	glutAddMenuEntry("Hollow"     , MENU_SELECT_HOLLOW_TREE);
	glutAddMenuEntry("Helix"      , MENU_SELECT_HELIX_TREE);
	glutAddMenuEntry("ConvexHelix", MENU_SELECT_SEG_HELIX_TREE);
	glutAddMenuEntry("SwissCheese", MENU_SELECT_SWISS_CHEESE_TREE);
	glutAddMenuEntry("HollowPipe" , MENU_SELECT_PIPE_TREE);

	toolsSubMenu = glutCreateMenu(menu);
	glutAddMenuEntry("Preview"           ,MENU_DRAW_PREVIEW);
	glutAddMenuEntry("DepthRange"		 ,MENU_DRAW_DEPTH_RANGE);
	glutAddMenuEntry("Wireframe"         ,MENU_DRAW_WIREFRAME);
	glutAddMenuEntry("Convexity"         ,MENU_DRAW_COUNT_SURFACES);
	glutAddMenuEntry("Goldfeather"       ,MENU_DRAW_GOLDFEATHER);
	glutAddMenuEntry("GoldfeatherLayers" ,MENU_DRAW_GOLDFEATHER_LAYER);
	glutAddMenuEntry("SCS"               ,MENU_DRAW_SCS);

	glutAddMenuEntry("Kmap"      ,MENU_DRAW_KMAP);

	configSubMenu = glutCreateMenu(menu);
	glutAddMenuEntry("Info"     , MENU_TOGGLE_INFO);
	glutAddMenuEntry("FrameRate", MENU_TOGGLE_FRAMERATE);
	glutAddMenuEntry("Copy/Read", MENU_TOGGLE_COPY);
	glutAddMenuEntry("Capture",   MENU_TOGGLE_CAPTURE);
	glutAddMenuEntry("WritePOV",  MENU_WRITE_POVRAY_OUTPUT);

	benchmarkSubMenu = glutCreateMenu(menu);
	glutAddMenuEntry("10"       , MENU_BENCHMARK_10);
	glutAddMenuEntry("100"      , MENU_BENCHMARK_100);
	glutAddMenuEntry("1000"     , MENU_BENCHMARK_1000);

	animateSubMenu = glutCreateMenu(menu);
	glutAddMenuEntry("None"     , MENU_ANIMATE_MODE_NONE);
	glutAddMenuEntry("RealTime" , MENU_ANIMATE_MODE_WALL_CLOCK);
	glutAddMenuEntry("FrameTime", MENU_ANIMATE_MODE_FRAME);

	makeMainMenu();

	glutMainLoop();

	return 0;
}

