#include "GeneticAlgorithm.h"
#include "TestString.h"

#include <math.h>
#include <assert.h>

#include <iomanip>
#include <map>

#define MIN _MIN
#define MAX _MAX 

bool verboseInfo = false;

//
// GENETIC ALGORITHM COMPONENTS
//

//
// Test string for fitness
//
// This function will return a value between
// 0 and 1 where 1 is perfectly fit.
//

double geneticFitness(const OString &str,const vector<OString> &perms)
{
	ulong match = 0;

	for (ulong p=0; p<perms.size(); p++)
		if (str.insideByDecimation(perms[p]))
			match++;

	return (double) match / (double) perms.size();
}

//
// Mutate a string by changing a 
// character at random
//

OString geneticMutate1(const OString &str)
{
	uint pos = rand()%str.size();
	uint offset = rand()%str.base();

	OString out = str;
	out[pos] = (str[pos]+offset)%out.base();

	return out;
}

//
// Mutate a string by swapping 
// two characters at random
//

OString geneticMutate2(const OString &str)
{
	uint pos1 = rand()%str.size();
	uint pos2 = rand()%str.size();

	OString out = str;
	uint tmp  = out[pos1];
	out[pos1] = out[pos2];
	out[pos2] = tmp;

	return out;
}

//
// Mutate a string by cyclicly
// rotating.
//

OString geneticMutate3(const OString &str)
{
	uint l    = str.size();
	uint pos = rand()%l;

	OString out(str.base(),l);
	memcpy(out.begin()       ,str.begin()+pos,(l-pos)*sizeof(ulong));
	memcpy(out.begin()+l-pos ,str.begin()    ,   pos *sizeof(ulong));

	return out;
}

//
// Perform reproduction on two strings
// by cutting somewhere.
//

OString geneticReproduction(const OString &a,const OString &b)
{
	long seed = rand();
	uint l = a.size();

	assert(l);

	OString aStr = (seed&8) ? a : a.mirror();
	OString bStr = (seed&4) ? b : b.mirror();

	uint pos = rand()%l;

	OString out(a.base(),l);

	memcpy(out.begin()    ,aStr.begin()    ,   pos *sizeof(ulong));
	memcpy(out.begin()+pos,bStr.begin()+pos,(l-pos)*sizeof(ulong));

	return out;
}

//
//
//

typedef pair<OString,double> OStringFitness;

double pushBackFitness(const vector<OString> &perms,vector<OStringFitness> &pop,const OString &str)
{
	double fitness = geneticFitness(str,perms);
	pop.push_back(OStringFitness(str,fitness));
	return fitness;
}

bool operator!=(const OString &str,const vector<OStringFitness> &pop)
{
	for (uint i=0; i<pop.size(); i++)
		if (str==pop[i].first)
			return false;

	return true;
}

void geneticAlgorithm(const uint n,const uint l,const uint p)
{
	uint pop = p;

	vector<OStringFitness> population;

#ifdef _DEBUG
	cout << "Populating world..." << endl;
#endif

	uint i;

	for (i=0; i<pop; i++)
		pushBackFitness(testAgainst,population,OString::randomNoDup(n,l));

	ulong gen = 0;
	OString found;

	for (;;)
	{
		gen++;

		double min = 1.0;  uint best = 0;
		double max = 0.0;  
		double mean = 0.0;
		ulong  good = 0;

		for (i=0; i<population.size(); i++)
		{
			double fitness = population[i].second;

			min = fitness<min ? fitness : min;

			if (fitness>max)
			{
				max = fitness;
				best = i;
				good = 1;
			}
			else
				if (fitness==max)
				{
					good++;
					best = i;
				}

			mean += fitness;
		}

		mean = mean / (double) population.size();

		if (verboseInfo)
		{
			cout << endl;
			cout << "Generation   -> " << gen << endl;
			cout << "Population   -> " << population.size() << endl;
			cout << "Min,Mean,Max -> " << min << ' ' << mean << ' ' << max << endl;
			cout << "Best         -> " << population[best].first << endl;
		}
		else
		{
			cout << setw(5) << gen << ' ' << population.size() << ' ';
			cout.setf(ios::fixed);
			cout << setprecision(5) << min << ' ' << mean << ' ';
			cout << setprecision(5) << max << ' ';
			cout << population[best].first << ' ' << good << endl;
		}

		if (max==1.0)
		{
			found = population[best].first;
			break;
		}

		// Next generation

		vector<OStringFitness> nextGen;

//		if (gen%25==0)
		if ((good<<1)>population.size())
		{
			// Get rid of duplicates every 10 generations

			for (i=0; i<population.size(); i++)
				if (population[i].first!=nextGen)
					nextGen.push_back(population[i]);
		}
		else
		{
			// Cull half population most years

			for (i=0; i<population.size(); i++)
				if (population[i].second>=mean)
					nextGen.push_back(population[i]);
		}

		if (!nextGen.size())
			nextGen.push_back(population[0]);
		else
			if (nextGen.size()==population.size())
				pop += pop/5;
//				nextGen.erase(nextGen.begin()+1,nextGen.end());

		for (i=nextGen.size(); i<pop; i++)
		{
			uint c = rand()%100;

			uint r = rand();
			#define r1 (r%i)
			#define r2 ((r/i)%i)

			if (c<20)
			{
				pushBackFitness
				(
					testAgainst,
					nextGen,
					geneticMutate1(nextGen[r1].first)
				);
				continue;
			}

			if (c<40)
			{
				pushBackFitness
				(
					testAgainst,
					nextGen,
					geneticMutate2(nextGen[r1].first)
				);
				continue;
			}

			if (c<50)
			{
				pushBackFitness
				(
					testAgainst,
					nextGen,
					geneticMutate3(nextGen[r1].first)
				);
				continue;
			}

			if (c<80)
			{
				pushBackFitness
				(
					testAgainst,
					nextGen,
					geneticReproduction(
						nextGen[r1].first,
						nextGen[r2].first)
				);
				continue;
			}
			
			if (c<95)
			{
				pushBackFitness
				(
					testAgainst,
					nextGen,
					OString::randomNoDup(n,l)
				);
				continue;
			}

			pushBackFitness
			(
				testAgainst,
				nextGen,
				nextGen[r1].first.normalise()
			);
		}

		population = nextGen;
	}

	testAccept.push_back(found);

	cout << endl << found << endl;
}

