#include "OString.h"

#include <algorithm>
#include <fstream>

//
// Constructors
//

bool OString::asAlpha = true;

OString::OString(const ulong n,const ulong l)
: _n(n)
{
	reserve(l+n);
	insert(begin(),l,0);
}

OString::OString(const char *str)
: _n(0)
{
	if (!str)
		return;

	const char &firstChar = *str;
	bool alpha = (firstChar>='a' && firstChar<='z');

	const unsigned char *pos = (const unsigned char *) str;

	while (*pos!=0)
	{
		ulong c = (ulong) (alpha ? (*pos)-'a' : (*pos)-'0');
		if (c+1>_n)
			_n = c+1;
		push_back(c);
		pos++;
	}
}

OString::OString(istream &is)
: _n(0)
{
	read(is);
}

ulong 
OString::base() const
{
	return _n;
}

//
// IO
//

istream &
OString::read(istream &is)
{
	ulong n,l;

	is.read((char *) &n,sizeof(ulong));
	is.read((char *) &l,sizeof(ulong));

	_n = n;
	erase(begin(),end());

	ulong *str = new ulong[l];
	is.read((char *) str,sizeof(ulong)*l);
	insert(begin(),str,str+l);
	delete [] str;

	return is;
}

ostream &
OString::write(ostream &os) const
{
	os.write((char *) &_n,sizeof(ulong));
	ulong l = size();
	os.write((char *) &l,sizeof(ulong));
	os.write((char *) &((*this)[(uint) 0]),sizeof(ulong)*l);
	return os;
}

istream &
OString::read(istream &is,vector<OString> &str)
{
	str.erase(str.begin(),str.end());

	for (;;)
	{
		OString s(is);
		if (!is.eof())
			str.push_back(s);
		else
			break;
	}	

	#ifdef _DEBUG
	cout << str.size() << " strings read." << endl;
	#endif

	return is;
}

ostream &
OString::write(ostream &os,const vector<OString> &str)
{
	for (ulong s=0; s<str.size(); s++)
		str[s].write(os);

	#ifdef _DEBUG
	cout << str.size() << " strings written." << endl;
	#endif

	return os;
}

//
// Convert OString to displayable string
//
// REVISIT - resolve display for large n
//

OString::operator string() const 
{
	string str("");

	for (uint i=0; i<size(); i++)
		if (asAlpha)
			str += 'a'+(char)((*this)[i]);
		else
			str += '0'+(char)((*this)[i]);

	return str;
}

ostream &
operator<<(ostream &os,const OString &str)
{
	return os << (string) str;
}

//
// Find next/previous string, lexigraphically.
//
// Return true if size of string has changed.
//

bool
OString::operator++()
{
	for (iterator i=end() - 1;;i--)
	{
		if ( ++(*i) < _n )
			return false;

		*i = 0;

		if (i==begin())
			break;
	}

	insert(begin(),1,0);

	return true;
}

bool
OString::operator--()
{
	for (iterator i=end() - 1;;i--)
	{
		if ( --(*i) < _n )
			return false;

		*i = _n - 1;

		if (i==begin())
			break;
	}

	erase(begin());

	return true;
}

// Comparison

bool 
OString::operator==(const OString &str) const
{
	if (_n!=str._n) return false;
	if (size()!=str.size()) return false;

	return !memcmp(begin(),str.begin(),size()*sizeof(ulong));
}

bool 
OString::operator!=(const OString &str) const
{
	return !((*this)==str);
}

bool 
OString::operator==(const vector<OString> &str) const
{
	for (ulong s=0; s<str.size(); s++)
		if ((*this)==str[s])
			return true;

	return false;
}

bool 
OString::operator!=(const vector<OString> &str) const
{
	return !((*this)==str);
}

//
// Catenation
//

OString  
OString::operator+(const OString &b)
{
	OString out(*this);
	out.insert(out.end(),b.begin(),b.end());
	return out;
}

OString &
OString::operator+=(const OString &b)
{
	insert(end(),b.begin(),b.end());
	return *this;
}

// Offset each element

OString  
OString::operator+ (const ulong &x)
{
	OString out(*this);
	for (uint i=0; i<out.size(); i++)
		out[i] = (out[i]+x)%_n;
	return out;
}

OString &
OString::operator+=(const ulong &x)
{
	_n += x;			// REVISIT

	for (uint i=0; i<size(); i++)
		(*this)[i] = ((*this)[i]+x)%_n;

	return *this;
}

//
//

bool
OString::insideByDecimation(const OString &str) const
{
	unsigned int tPos=0;
	unsigned int sPos=0;

	for (;;)
	{
		if ((*this)[tPos]==str[sPos])
			sPos++;
		tPos++;

		if (sPos==str.size())
			return true;

		if (tPos==size())
			return false;
	}	
}

OString
OString::normalise() const
{
	OString trans(_n);
	trans.insert(trans.begin(),_n,'.');

	OString out(_n);
	ulong next = 0;

	for (uint i=0; i<size(); i++)
	{
		uint ch = (*this)[i];

		if (trans[ch]=='.')
			trans[ch] = next++;

		out.push_back(trans[ch]);
	}

	return out;
}

// Mirror the string

OString 
OString::mirror() const
{
	OString out(_n);

	for (uint i=0; i<size(); i++)
		out.push_back((*this)[size()-1-i]);
	
	return out;
}

OString 
OString::skipNth(const uint pos) const
{
	OString out(_n);

	for (uint i=0; i<size(); i++)
		if (i!=pos)
			out.push_back((*this)[i]);

	return out;
}

OString 
OString::random(const uint n,const uint l)
{
	OString out(n);

	for (uint p=0; p<l; p++)
		out.push_back(rand()%n);

	return out;
}

OString
OString::randomNoDup(const uint n,const uint l)
{
	OString out(n,l);

	ulong last = rand()%n;
	out[0] = last;

	for (uint p=1; p<l; p++)
	{
		out[p] = (last + 1 + rand()%(n-1))%n;
		last = out[p];
	}

	return out;
}

OString 
OString::permutation(const ulong n,const ulong p,const ulong k)
{
	//
	// Check that k is not too large
	//

	assert(k<=n);

	if (k>n)
		return OString();

	//
	// Do it
	// 

	ulong length = k ? k : n;

	//
	// Create 'abcde...'
	//

	OString perm(n);
	for (ulong c=0; c<n; c++)
		perm.push_back(n-1-c);

	//
	//
	//

	OString out(n);
	ulong seed = p;

	while (out.size()<length)
	{
		ulong choices = perm.size();
		ulong select = seed%choices;
		seed = seed/choices;

		out.insert(out.begin(),perm[(uint) select]);
		perm.erase(perm.begin()+select);
	}

	return out;
}

vector<OString> 
OString::permutations(const ulong n,const ulong k)
{
	//
	// Default behaviour is when k
	// is assumed to be n
	//
	// k is the 'length' of desired
	// permutations.  k!=n implies
	// 'arrangements'.
	//

	if (k==0 || k==n)
	{
		ulong nf = factorial(n);
		vector<OString> perms;
		perms.reserve(nf);

		for (ulong p=0; p<nf; p++)
			perms.push_back(permutation(n,p));

		return perms;
	}
	else

	// 
	// Do every arrangement of n 
	// select k.
	//

	{
		ulong arr = factorial(n)/factorial(n-k);
		vector<OString> perms;
		perms.reserve(arr);

		for (ulong p=0; p<arr; p++)
			perms.push_back(permutation(n,p,k));

		return perms;
	}
}

OString
OString::transposition(const uint a,const uint b) const
{
	OString out = *this;
	
	assert(out.size());
	assert(a<out.size());
	assert(b<out.size());

	ulong tmp = out[a];
	out[a] = out[b];
	out[b] = tmp;

	return out;
}

//
//

OString 
OString::substitute(const OString s) const
{
	OString out(_n);

	//
	// Check that remapping is consistent with
	// size of alphabet
	//

	assert(s._n<=_n);

	if (s._n>_n)
		return out;

	//
	// Check that remapping caters for every
	// character in alphabet.
	//

	assert(_n==s.size());

	if (_n!=s.size())
		return out;

	// Do it

	for (ulong c=0; c<size(); c++)
		out.push_back(s[(*this)[c]]);
	
	return out;
}

//
//

ulong
factorial(const ulong n)
{
	ulong nFactorial = 1;
	
	for (ulong f=n; f>1; f--)
		nFactorial *= f;

	return nFactorial;
}

ulong
power(const ulong a,const ulong b)
{
	if (b==0)
		return 1;

	ulong pow = 1;
	for (ulong i=0;i<b;i++)
		pow *= a;

	return pow;
}



ostream &operator<<(ostream &os,vector<OString> &str)
{
	for (uint s=0; s<str.size(); s++)
		os << str[s] << endl;

	return os;
}

void displayString(const OString &str)
{ 
	cout << (string) str << endl; 
}

//
//
//

