//
// PSSU (Permutation String Search Utility)
// Nigel Stewart (nigels@eisa.net.au)
//

#include <OString.h>

#include <TestString.h>

#include <Normalised.h>
#include <AllGen.h>
#include <NikGen.h>
#include <RobGen.h>
#include <GeneticAlg.h>
#include <AltGroup.h>

#include <string>
#include <fstream>

#include <algorithm>

#include <time.h>

//

ulong n = 0;
ulong k = 0;	
ulong l = 0;	// Length of search space
ulong p = 0;	// Population size
ulong s = 0;	// Random seed

bool histogram = false;
bool searchL = true;

string inFile;
bool   outFile = false;
string criteriaFile;
string gifFile;

time_t startTime;
time_t endTime;
time_t elapsedTime;

//
// Support different algorithms
//

typedef enum PermuteAlgorithm
{
	SEARCH_ALL = 0,
	SEARCH_NORMALISED,
	SEARCH_NIK_LYGEROS,
	SEARCH_ROBERT_ERRA,
	SEARCH_GENETIC_ALGORITHM,
	SEARCH_UNIQUE_SOLUTIONS,
	SEARCH_TYPES
} PermuteAlgorithm;

PermuteAlgorithm searchAlg = SEARCH_NORMALISED;

char *PermuteAlgorithmDescription[SEARCH_TYPES] =
{
	"All",
	"Normalised",
	"Nik Lygeros",
	"Robert Erra",
	"Genetic Algorithm",
	"Unique Solutions"
};

//
// Support different criteria
//

typedef enum DecimationCriteria
{
	CRITERIA_ALL_PERMUTATIONS = 0,
	CRITERIA_EVEN_ALTERNATING_GROUP,
	CRITERIA_ODD_ALTERNATING_GROUP,
	CRITERIA_TYPES
} DecimationCriteria;

DecimationCriteria decCriteria = CRITERIA_ALL_PERMUTATIONS;

char *DecimationCriteriaDescription[CRITERIA_TYPES] =
{
	"All Permutations",
	"Even Alt. Group",
	"Odd Alt. Group"
};

//
// Software info
//

char softwareTitle[] =
	"Permutation Sequence Search Utility \n" \
	"Nigel Stewart (nigels@eisa.net.au)  \n" \
	"  Incorporating GAlib C++ Genetic Algorithm Library \n" \
	"  Copyright (c) 1994-1996 Massachusetts Institute of Technology\n\n"; 

char softwareParameters[] =
	"n=x         Number of objects.		                \n" \
	"l=x         Length of strings in search space.     \n" \
	"k=x         Length of permutations required.       \n" \
	"p=x         Population size for genetic algorithm. \n" \
	"odd         Test against odd alternating group.    \n" \
	"even        Test against odd alternating group.    \n" \
	"-o          Write satisfactory sequences to file.  \n" \
	"-cfilename  Write criteria to file.                \n" \
	"nik         Use Nik's generating algorithm \n" \
	"robert      Use Robert Erra's mirror algorithm \n" \
	"ga          Use genetic algorithm \n" \
	"unique      Use search-for-unique algorithm \n" \
	"search      Increment l until successfull.	\n" \
	"histogram   Count number of permutations.      \n";

char softwareHelpScreen[] =
	"Usage                                          \n" \
	"---------------------------------------------- \n" \
 	"                                               \n" \
	"pssu -imydata.dat                              \n" \
    "             Read strings from file and display\n" \
 	"                                               \n" \
	"pssu n=4 [-opermute4.dat]                      \n" \
    "             Find 4! permutations of a,b,c,d   \n" \
    "             Optionally, write to file         \n" \
 	"                                               \n" \
	"pssu n=4 l=6 [-ofile.dat] [search] [histogram] \n" \
    "             Search for strings containing 4!  \n" \
	"             permutations, starting with       \n" \
    "             strings of length 6.  \"search\"    \n" \
    "             will increment the length until   \n" \
	"             a match is found.  \"histogram\"    \n" \
	"             will count the number of          \n" \
	"             permutations in each test string. \n";

//

void printConfig(ostream &os,int argc, char *argv[]);
void printSummary(ostream &os);

//
//

int main(int argc,char *argv[])
{
	for (int arg=1; arg<argc; arg++)
	{
		string a(argv[arg]);

		if (!strncmp(a.c_str(),"n=",2))			// n
			n = (ulong) atoi(a.c_str()+2);
		else if (!strncmp(a.c_str(),"l=",2))	// l
			l = (ulong) atoi(a.c_str()+2);
		else if (!strncmp(a.c_str(),"k=",2))	// k
			k = (ulong) atoi(a.c_str()+2);
		else if (!strncmp(a.c_str(),"p=",2))
			p = (ulong) atoi(a.c_str()+2);
		else if (a=="-o")						// output filename
			outFile = true;
		else if (!strncmp(a.c_str(),"-i",2))	// input filename
			inFile = string(a.c_str()+2);
		else if (!strncmp(a.c_str(),"-c",2))
			criteriaFile = string(a.c_str()+2);
		else if (!strncmp(a.c_str(),"-g",2))	// gif filename
			gifFile = string(a.c_str()+2);
		else if (a=="histogram")				// histogram
			histogram = true;
		else if (a=="search")					// search
			searchL = true;
		else if (a=="nik")
			searchAlg = SEARCH_NIK_LYGEROS;
		else if (a=="robert")
			searchAlg = SEARCH_ROBERT_ERRA;
		else if (a=="ga")
			searchAlg = SEARCH_GENETIC_ALGORITHM;
		else if (a=="all")
			searchAlg = SEARCH_ALL;
		else if (a=="unique")
			searchAlg = SEARCH_UNIQUE_SOLUTIONS;
//		else if (a=="notnorm")
//			testNormalise = false;
		else if (a=="odd")
			decCriteria = CRITERIA_ODD_ALTERNATING_GROUP;
		else if (a=="even")
			decCriteria = CRITERIA_EVEN_ALTERNATING_GROUP;
		else if (!strncmp(a.c_str(),"-h",2))	// help
		{
			cout << softwareTitle << endl << softwareHelpScreen << endl; 
			exit(0);
		}
		else
		{
			cout << argv[0] << ": " << a << " (Invalid parameter) " << endl;
			exit(0);
		}
	}

	if (!inFile.empty())
	{
		vector<OString> str;
#ifdef __CYGWIN32__
		ifstream inF(inFile,ios::in|ios::binary);
#else
		ifstream inF(inFile.c_str(),ios::in);
#endif
		OString::read(inF,str);
		cout << str;
		return 0;
	}

	if (n==0 || n==0 && l==0 && k==0)
	{
		cout << softwareTitle << endl << softwareParameters << endl;
		exit(0);
	}

	if (!s)
		s = time(NULL);

	srand(s);

	printConfig(cout,argc,argv);

	// If k not specified, assume n

	if (!k)
		k = n;

	// If length not specified, try length n

	if (!l)
		l = n;

	// If p not specified, assume 100

	if (!p)
		p = 100;

	// Permutations

/*	testAccept = vector<OString>();  */

	StringTest test(STRING_TEST_ALL_BY_DECIMATION,histogram);

	switch (decCriteria)
	{
	default:
	case CRITERIA_ALL_PERMUTATIONS:
		{
			#ifdef _DEBUG
			cout << "Finding permutations for n=" << n << endl;
			#endif

			test.criteria() = OString::permutations(n,k);
		}
		break;
	case CRITERIA_ODD_ALTERNATING_GROUP:
		{
			#ifdef _DEBUG
			cout << "Finding odd alternating group for n=" << n << endl;
			#endif

			alternatingGroups(n,vector<OString>(),test.criteria());
		}
		break;
	case CRITERIA_EVEN_ALTERNATING_GROUP:
		{
			#ifdef _DEBUG
			cout << "Finding even alternating group for n=" << n << endl;
			#endif

			alternatingGroups(n,test.criteria(),vector<OString>());
		}
		break;
	}

	cout << test.criteria().size() << " criteria created." <<  endl << endl;

	if (!criteriaFile.empty())
	{
#ifdef __CYGWIN32__
		ofstream outF(criteriaFile,ios::out|ios::binary);
#else
		ofstream outF(criteriaFile.c_str(),ios::out);
#endif
		outF << test.criteria();
		outF.close();

		return 0;
	}

	#ifdef _DEBUG
	cout << "Searching..." << endl;
	#endif

	if (l)
		for (;;)
		{
			cout << "l=" << l << ' ' << endl;

			test.reset();
			startTime = clock();

			switch (searchAlg)
			{
			case SEARCH_ALL:
				generateAll(n,l,test);
				break;
			case SEARCH_UNIQUE_SOLUTIONS:
				if (test.accepted().size())
				{
					StringTest test2(STRING_TEST_NONE_BY_DECIMATION);
					test2.criteria() = test.accepted();
					test2.moreCriteria().push_back(test);
					generateAll(n,l,test2);
					if (test2.accepted().size())
						return 1;
				}
				else
					generateAll(n,l,test);
				break;
			case SEARCH_NORMALISED:
				generateNormalised(n,l,test);
				break;
			case SEARCH_NIK_LYGEROS:
				nikLygeros(n,test);
				break;
			case SEARCH_ROBERT_ERRA:
				erraMirror(n,test);
				break;
			case SEARCH_GENETIC_ALGORITHM:
				geneticAlgorithm(n,k,l,p,test,s);
				break;
			}

			if (test.accepted().size() && outFile && searchAlg!=SEARCH_UNIQUE_SOLUTIONS)
			{
				char filename[1024];
				sprintf(filename,"outn%uk%ul%ut%u.txt",n,k,l,searchAlg);

#ifdef __CYGWIN32__
				ofstream outF(filename,ios::out|ios::binary);
#else
				ofstream outF(filename,ios::out);
#endif
				printConfig(outF,argc,argv);
				printSummary(outF);

				outF << test.accepted();

				outF.close();
			}

			cout << test.testCount() << " decimation tests" << endl;
			if (histogram)
			{
				cout << "Histogram..." << endl;
				test.histogram(cout);
			}

			cout << endl;

			if ((test.accepted().size() || !searchL) && searchAlg!=SEARCH_UNIQUE_SOLUTIONS)
				break;

			l++;
		}

	cout << endl;

	printSummary(cout);

	return true;
}

//
// Config info
//

void printConfig(ostream &os,int argc, char *argv[])
{
	os << '\"';	
	for (int a=0;a<argc;a++) os << argv[a] << ' ';
	os << '\"' << endl;

	os << endl;
	os << softwareTitle;
	os << "Binary (" << __DATE__ << ' ' << __TIME__ << ')' << endl;
	os << "n=" << n << ' ';
	os << "l=" << l << ' ';
	os << "k=" << k << ' ';
	os << "p=" << p << ' ';
	os << "s=" << s << ' ';
	os << endl;
	os << "algorithm=\""<< PermuteAlgorithmDescription[searchAlg] << "\" ";
	os << endl;
	os << "criteria=\""<< DecimationCriteriaDescription[decCriteria] << "\" ";
	os << endl;
	os << endl;
}

//
// Summary info
//

void printSummary(ostream &os)
{
	endTime = clock();

	elapsedTime = endTime - startTime;

	const time_t sec = CLOCKS_PER_SEC;
	const time_t min = 60*CLOCKS_PER_SEC;
	const time_t hour = 60*60*CLOCKS_PER_SEC;

	os << "Elapsed Time (hr:min:sec) ";
	os << elapsedTime/hour << ':';
	os << (elapsedTime/min)%60 << ':';
	os << (elapsedTime%min)/(double) CLOCKS_PER_SEC << endl;

	os << StringTest::testCountTotal() << " decimation tests in total." << endl;
	os << endl;
}

