#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glut.h>
#include <atimer.h>
#include <stdlib.h>

//
// Benchmark of glCopyPixels for GL_COLOR,GL_STENCIL,GL_DEPTH
//
// System Setup:
//	Windows NT 3.51, Pentium 100, 32Meg RAM
//	3Demon SX88, Display Driver 2.0, OpenGL driver 1.0.9
//	Visual C++ V4.0
//
// Compiling:
//		Required libraries: opengl32.lib glu32.lib glaux.lib
//
// Sample Output:
//
// (500x500)
// Color rate  : 1.5625e+007
// Depth rate  : 3.90625e+006
// Stencil rate: 3.90625e+006
//
// (1024x745)
// Color rate  : 1.12188e+007
// Depth rate  : 3.97333e+006
// Stencil rate: 4.04926e+006
//
// (554x115)
// Color rate  : 1.59275e+007
// Depth rate  : 3.98187e+006
// Stencil rate: 3.98187e+006
//
// Prepared by Nigel Stewart, RMIT University, 10-01-1999
// Email: nigels@eisa.net.au
//
//

const GLubyte Red[3] =  { 255,  0,   0};
const GLubyte Blue[3] = {   0,  0, 255};

GLdouble maximumSamplePeriod = 0.3F;
const GLdouble Mill = 1.0/1000000.0;

void Draw(void)
{
	GLuint viewport[4];
	GLuint width,height;

	glGetIntegerv(GL_VIEWPORT,(GLint *) viewport);
	width = viewport[2];
	height = viewport[3];

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );

	glViewport(0,0,width>>1,height>>1);
	glColor3ubv(Red);
	glDepthFunc(GL_LESS);
	glutSolidSphere(1.0,10,10);

	glViewport(0,0,width,height);
	glRasterPos2i(0,0);

	cout << width << '\t' << height << '\t';

	{
		GLuint repetitions = 0;
		AcTimer timer(cout);

		while (1)
		{
			glCopyPixels(0,0,width>>1,height>>1,GL_COLOR);
			glFinish();

			repetitions++;

			if (timer.Elapsed()>maximumSamplePeriod) break;
		}

		GLdouble rate = 0.25*repetitions*width*height/timer.Elapsed();
		cout << repetitions << " ";
		cout << rate*Mill << "M \t";
	}

	{
		GLuint repetitions = 0;
		AcTimer timer(cout);

		glColor3ubv(Blue);
		glDepthFunc(GL_NOTEQUAL);

		while (1)
		{
			glCopyPixels(0,0,width>>1,height>>1,GL_DEPTH);
			glFinish();

			repetitions++;

			if (timer.Elapsed()>maximumSamplePeriod) break;
		}

		GLdouble rate = 0.25*repetitions*width*height/timer.Elapsed();
		cout << repetitions << " ";
		cout << rate*Mill << "M \t";
	}

	{
		GLuint repetitions = 0;
		AcTimer timer(cout);

		while (1)
		{
			glCopyPixels(0,0,width>>1,height>>1,GL_STENCIL);
			glFinish();

			repetitions++;

			if (timer.Elapsed()>maximumSamplePeriod) break;
		}

		GLdouble rate = 0.25*repetitions*width*height/timer.Elapsed();
		cout << repetitions << " ";
		cout << rate*Mill << 'M' << endl;
	}

	glutSwapBuffers();

}

void Reshape(int width, int height)
{
	glViewport(0, 0, width, height);
}

void Key(unsigned char key,int x,int y)
{
	exit(0);
}

int main(int argc,char *argv[])
{
	glutInit(&argc,argv);
	glutInitWindowSize(250,250);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STENCIL);
	glutCreateWindow("glCopyPixel Speed");

	glEnable(GL_DEPTH_TEST);
	glClearDepth(1.0);

	glTranslatef(0.0,0.0,-1.0);

	glutKeyboardFunc(Key);
	glutReshapeFunc(Reshape);
	glutDisplayFunc(Draw);

	cout << "          Vendor : " << (char *) glGetString(GL_VENDOR) << endl;
	cout << "        Renderer : " << (char *) glGetString(GL_RENDERER) << endl;
	cout << "         Version : " << (char *) glGetString(GL_VERSION) << endl;

	GLint zDepth;
	GLint stencilDepth;
	GLint doubleBuffering;
	GLint rgbaMode;

	glGetIntegerv(GL_DEPTH_BITS,&zDepth);
	glGetIntegerv(GL_STENCIL_BITS,&stencilDepth);
	glGetIntegerv(GL_DOUBLEBUFFER,&doubleBuffering);
	glGetIntegerv(GL_RGBA_MODE,&rgbaMode);


	cout << "Double-buffering : " << (doubleBuffering ? "" : "not " ) << "supported." << endl;
	cout << "       RGBA mode : " << (rgbaMode ? "" : "not " ) << "enabled." << endl;
	cout << "    Depth buffer : " << zDepth << " bit-planes." << endl;
	cout << "  Stencil buffer : " << stencilDepth << " bit-planes." << endl;

	cout << "\nwidth\theight\tcolour\t\tdepth\t\tstencil" << endl;

	glutMainLoop();
	
	return 0;
}
