//
// OpenGL Image Space Surface Counting
//
// This routine assumes that OpenGL has been 
// setup with the desired triangle culling.
//
// Nigel Stewart (nts@minyos.its.rmit.edu.au)
// RMIT Department of Manufacturing Systems Engineering
// 13th December 1996
//

#include <cntsurf.h>

#include <zplane.h>
#include <iostream.h>

////////////////////////////////////////////////////////////
//
// 1-bit stencil work-around.
//
// Define FRAME_DATABASE to use it, 
// 0 to enable collection, 1 to enable 
// utilisation.
//
// This allows SCS to operate on hardware
// with one-bit stencil buffer, by pre-computing
// k for various viewing directions.
// It does not solve layer extraction requirement
// of Goldfeather and Layered Goldfeather
//
// Also see file src-csg/convexity.cpp
// 

//#define FRAME_DATABASE 1

#ifdef FRAME_DATABASE
#include <atypes.h>
typedef Uint convexTable[1000][2];
convexTable *CurrentConvexTable = NULL;
extern Uint frameCounter;
#endif

////////////////////////////////////////////////////////////

GLubyte
countSurfaces(GLuint *obj,const GLuint objs,const countSurfacesMode mode)
{
	glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

		// Ensure stencil buffer is fully zero

		glClearStencil(0x00);
		glStencilMask(0xff);
		glClear(GL_STENCIL_BUFFER_BIT);
		
		// Disable z buffer test & update
		
		glDepthMask(GL_FALSE);
		glDepthFunc(GL_ALWAYS);

		// Use stencil test to count surfaces

		glEnable(GL_STENCIL_TEST);
		glStencilFunc(GL_ALWAYS,0xff,0xff);
		glStencilOp(GL_INCR,GL_INCR,GL_INCR);

		// Don't bother with drawing colour

		glDisable(GL_LIGHTING);
		glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);

		// Rasterise objects

		for (GLuint o=0;o<objs;++o)
			glCallList(obj[o]);

		// Get histogram
		
		GLuint histogram[256];
		GLuint pixels = stencilHistogram(histogram);

		GLubyte max = 0;

		switch (mode)
		{

		case COUNT_SURFACES_ALL:
			{
				for (GLuint p=0; p<256; p++)
					if (histogram[p]>0)
						max = p;
			}
			break;

		case COUNT_SURFACES_99TH_PERCENTILE:
			{
				GLuint accum = 0;
				for (GLuint p=0; p<256; p++)
				{
					accum += histogram[p];
					if ((double) accum / pixels >= 0.999)
					{
						max = p;
						break;
					}
				}
			}
			break;
		}

	glPopAttrib();

	#ifdef _DEBUG
	GLint stencilSize = 0;
	glGetIntegerv(GL_STENCIL_BITS,&stencilSize);

	if ( GLubyte(GLubyte(1)<<stencilSize) == max+1 )
		cout << "WARNING: Stencil Buffer may be too small for countSurfaces" << endl;
	#endif

	#if (defined(FRAME_DATABASE) && (FRAME_DATABASE==0))
	cout << "  { " << frameCounter << ", " << Uint(max) << " }," << endl;
	#endif

	#if (defined(FRAME_DATABASE) && (FRAME_DATABASE==1))
	if (convexTable)
	{
		max = (*CurrentConvexTable)[frameCounter][1];

		  #ifdef _DEBUG
		  cout << "Look-up k = " << Uint(max) << endl;
          #endif
	}
	#endif

	return max;
}

//
// OpenGL Surface Counting Visualisation
//
// This routine assumes that OpenGL has been 
// setup with the desired triangle culling.
//

#define maxColors 7
 
GLubyte surfaceColor[maxColors][3] =
{
	{ 255,  0,  0 },	// Red
	{   0,255,  0 },	// Green
	{   0,  0,255 },	// Blue
	{ 255,255,  0 },	
	{ 255,  0,255 },
	{   0,255,255 },
	{ 255,255,255 }
};

GLubyte
countSurfacesVisualise(GLuint *obj,const GLuint objs)
{
	glPushAttrib( GL_ENABLE_BIT | GL_STENCIL_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT );

		GLubyte layers = countSurfaces(obj,objs,COUNT_SURFACES_ALL);

		glEnable(GL_STENCIL_TEST);
		glStencilMask(0xff);
		glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);

		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_ALWAYS);
		glDepthMask(GL_FALSE);

		glDisable(GL_LIGHTING);

		glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);

		for (GLuint l=0; l<layers && l<maxColors; l++)
		{
			if (l<layers)
			{
				glStencilFunc(GL_EQUAL,l+1,0xff);
				glColor3ubv(surfaceColor[l%(maxColors-1)]);
			}
			else
			{
				glStencilFunc(GL_GEQUAL,l+1,0xff);
				glColor3ubv(surfaceColor[maxColors-1]);
			}

			drawZnear();
		}

	glPopAttrib();

	return layers;
}

//
// OpenGL Stencil Buffer Histogram
//

GLuint
stencilHistogram(GLuint *histogram)
{
	GLuint viewport[4];
	glGetIntegerv(GL_VIEWPORT,(GLint *) viewport);

	GLuint   pixels = viewport[2] * viewport[3];
	GLubyte *pixel  = new GLubyte[pixels];

	glPixelStorei(GL_PACK_ALIGNMENT,sizeof(GLubyte));
	glReadPixels(0,0,viewport[2],viewport[3],GL_STENCIL_INDEX,GL_UNSIGNED_BYTE,pixel);

	// Initialise Histogram

	GLuint x;

	for (x=0;x<256;x++)
		histogram[x] = 0;

	// Build Histogram

	for (x=0;x<pixels;x++)
		histogram[pixel[x]]++;

	// Free allocated memory

	delete [] pixel;

	return pixels;
}
