//
// OpenGL Solid Modelling
//
// Nigel Stewart (nigels@eisa.net.au)
// RMIT Department of Mechanical and Manufacturing Engineering
// 3rd January 1999
//

#include <csgprim.h>
#include <sweep.h>
#include <func.h>

#include <string.h>
#include <assert.h>

CsgPrimitiveDatabase::CsgPrimitiveDatabase()
{
	// Sphere centered on origin, radius of 1.0

	GLuint sphere = glGenLists(1);
	glNewList(sphere,GL_COMPILE);
		glutSolidSphere(1.0,25,25);
	glEndList();

	add("sphere",sphere);

	// Unit box from <-1,-1,-1> to <1,1,1>

	GLuint box = glGenLists(1);
	glNewList(box,GL_COMPILE);
		glutSolidCube(2.0);
	glEndList();

	add("box",box);

	// Cylinder centered on origin, axis in z, radius of 1.0

	GLuint cylinder = glGenLists(1);
	glNewList(cylinder,GL_COMPILE);
	  glPushMatrix();
 		  glRotatef(90.0,0.0,1.0,0.0);
 		  glTranslatef(0.0,0.0,-1.0);
		  GLUquadricObj *quadObj = gluNewQuadric();
		  gluQuadricDrawStyle(quadObj, (GLenum) GLU_FILL);
		  gluQuadricNormals(quadObj, (GLenum) GLU_SMOOTH);
		  gluQuadricOrientation(quadObj, (GLenum) GLU_OUTSIDE);
		  gluCylinder(quadObj, 1.0, 1.0, 2.0, 15,20);
		  gluQuadricOrientation(quadObj, (GLenum) GLU_INSIDE);
		  gluDisk(quadObj,0.0,1.0,15,5);
		  glTranslatef(0.0,0.0,2.0);
		  glRotatef(180,0.0,1.0,0.0);
		  gluDisk(quadObj,0.0,1.0,15,5);
	  glPopMatrix();
	glEndList();

	add("cylinder",cylinder);

	// Helix

	const Uint samples = 16;
	const Uint divisions = 20;

	const Vector origin(-VectorX*3.5);
	const Vector direction(VectorX*7.0);
	const double radius = 0.6;

	Helix curve(origin,direction,1.0,2.0);
	GLuint helix = Sweep(curve,radius,divisions,samples);
	add("helix",helix);

	vector<Mesh *> meshes = MeshSweepConvex(curve,radius,divisions,samples);

	#ifdef _DEBUG 
	cout << meshes.size() << " segments." << endl;
	#endif

	for (Uint d=0; d<meshes.size(); d++)
	{
		char buffer[1024];
		sprintf(buffer,"HelixSegment%d",d);
		char *label = strdup(buffer);

		GLuint seg = glGenLists(1);
		glNewList(seg,GL_COMPILE);
		meshes[d]->glDraw();
		glEndList();
		add(label,seg);

//		meshes[d]->writePOV(cout);
	}
}

bool
CsgPrimitiveDatabase::add(char *name,const GLuint dl)
{
	#ifdef _DEBUG
	cout << "Adding " << name << " to display list database." << endl;
	#endif

	if (_displayLists[name])
	{
		assert(!_displayLists[name]);
		return false;
	}

	_displayLists[name] = dl;

	return true;
}

bool
CsgPrimitiveDatabase::add(char *name,CsgTreeLeaf *leaf)
{
	#ifdef _DEBUG
	cout << "Adding " << name << " to primitive database." << endl;
	#endif

	if (_namedPrimitives[name])
	{
		assert(!_namedPrimitives[name]);
		return false;
	}

	_namedPrimitives[name] = leaf;

	return true;
}

GLuint 
CsgPrimitiveDatabase::lookupDL(char *name)
{
	return _displayLists[name];
}

CsgTreeLeaf * 
CsgPrimitiveDatabase::lookupP(char *name)
{
	return _namedPrimitives[name];
}

ostream &
CsgPrimitiveDatabase::writePOV(ostream &os,const CsgTreeLeaf &leaf)
{
	bool found = false;

	if (lookupDL("sphere")==leaf._displayList)
		os << "sphere { 0 1 " << endl;
	else
	if (lookupDL("box")==leaf._displayList)
		os << "box { <-1,-1,-1>, <1,1,1> " << endl;
	else
	if (lookupDL("cylinder")==leaf._displayList)
		os << "cylinder { -x x 1.0 " << endl;
	else
		return os;

	os << "  "; leaf.writePOV(os) << endl;
	if (leaf.color())
	{
		os << "  pigment { rgb <";
		os << leaf.color()[0] << ',';
		os << leaf.color()[1] << ',';
		os << leaf.color()[2];
		os << "> } " << endl;
	}
	os << "}" << endl;
	return os;
}

ostream &
CsgPrimitiveDatabase::writePOV(ostream &os,const CsgTreeNode &tree)
{
	switch (tree.type())
	{
	case PRIMITIVE_NODE:
		writePOV(os,*((CsgTreeLeaf *) (&tree)));
		break;

	case UNION_NODE:
		os << "union { " << endl;
		writePOV(os,*(tree.left()));
		writePOV(os,*(tree.right()));
		os << "}" << endl;
		break;

	case INTERSECTION_NODE:
		os << "intersection { " << endl;
		writePOV(os,*(tree.left()));
		writePOV(os,*(tree.right()));
		os << "}" << endl;
		break;

	case DIFFERENCE_NODE:
		os << "difference { " << endl;
		writePOV(os,*(tree.left()));
		writePOV(os,*(tree.right()));
		os << "}" << endl;
		break;
	}

	return os;
}
