//
// OpenGL Solid Modelling
//
// Nigel Stewart (nts@minyos.its.rmit.edu.au)
// RMIT Department of Mechanical and Manufacturing Engineering
// 10th September 1997
//

#include <csgtree.h>
#include <zplane.h>

#include <string.h>
#include <assert.h>

/////////////////////
// CSG Product
//

CsgProduct::CsgProduct()
{
	_uncomplemented.reserve(1);
	_complemented.reserve(1);
}

CsgProduct::CsgProduct(const CsgProduct &prod)
: _uncomplemented(prod._uncomplemented),
  _complemented(prod._complemented)
{
}

CsgProduct &
CsgProduct::operator=(const CsgProduct &prod)
{
	_uncomplemented = prod._uncomplemented;
	_complemented = prod._complemented;

	return *this;
}

vector<CsgTreeLeaf *> &
CsgProduct::uncomplemented() 
{ 
	return _uncomplemented; 
}

vector<CsgTreeLeaf *> &
CsgProduct::complemented()   
{ 
	return _complemented; 
}

const vector<CsgTreeLeaf *> &
CsgProduct::uncomplemented() const
{ 
	return _uncomplemented; 
}

const vector<CsgTreeLeaf *> &
CsgProduct::complemented() const   
{ 
	return _complemented; 
}

int 
CsgProduct::size() const
{
	return _uncomplemented.size() + _complemented.size(); 
}

CsgTreeLeaf *
CsgProduct::element(const Uint i) const
{
	assert(i<size());

	if (i<_uncomplemented.size())
		return _uncomplemented[i];
	else
		return _complemented[i-_uncomplemented.size()];
}

bool 
CsgProduct::elementComplemented(const Uint i) const
{
	assert(i<size());

	return i>=_uncomplemented.size();
}

ostream &
operator<<(ostream &os,CsgProduct &prod)
{
	for (int p=0; p<prod._uncomplemented.size(); p++)
	{
		if (p!=0)
			os << '.';
		os << ((CsgTreeLeaf *) prod._uncomplemented[p])->name();
	}

	for (int c=0; c<prod._complemented.size(); c++)
	{
		os << '-';
		os << ((CsgTreeLeaf *) prod._complemented[c])->name();
	}

	return os;
}

////////////////////////
// CSG Sum of Products
//

CsgSumOfProducts::CsgSumOfProducts()
{
	reserve(1);
}

CsgSumOfProducts &
CsgSumOfProducts::operator=(const CsgSumOfProducts &sum)
{
	*((vector<CsgProduct> *)this) = sum;

	return *this;
}


ostream &
operator<<(ostream &os,CsgSumOfProducts &sumOfProd)
{
	for (int p=0; p<sumOfProd.size(); p++)
	{
		if (p>0)
			os << " + ";

		os << sumOfProd[p];
	}

	os << endl;

	return os;
}

////////////////////////
// CSG Node
//

list<CsgTreeNode *>   CsgTreeNode::_allocList;
bool                  CsgTreeNode::_allocListEnable(false);

void 
CsgTreeNode::startAllocList()
{
	_allocListEnable = true;
}

void 
CsgTreeNode::endAllocList()
{
	_allocListEnable = false;
}

void 
CsgTreeNode::deleteAllocList()
{
	while (_allocList.size())
		delete _allocList.front();
}

CsgTreeNode::CsgTreeNode(const CsgTreeNodeType type)
: _nodeType(type), _left(NULL), _right(NULL)
{
	if (_allocListEnable)
		_allocList.push_back(this);
}

CsgTreeNode::CsgTreeNode(const CsgTreeNode& node)
: _nodeType(node._nodeType), _left(node._left), _right(node._right)
{
	if (_allocListEnable)
		_allocList.push_back(this);

	if (_nodeType==PRIMITIVE_NODE)
		return;

	assert(_left);
	assert(_right);
	assert(_left!=_right || _left->_nodeType==PRIMITIVE_NODE);

	_left = _left->clone();
	_right = _right->clone();
}

CsgTreeNode::~CsgTreeNode()
{
	_allocList.remove(this);

	#ifdef _DEBUG
//	cout << "Calling delete for CsgTreeNode " << (void *) this << endl;
	#endif
}

CsgTreeNode *
CsgTreeNode::clone()
{
	if (type()==PRIMITIVE_NODE)
		return this;
	else
		return new CsgTreeNode(*this);
}

CsgTreeNodeType
CsgTreeNode::type() const
{
	return _nodeType;
}

CsgTreeNode *&
CsgTreeNode::left() 
{
	return _left;
}

CsgTreeNode *&
CsgTreeNode::right() 
{
	return _right;
}

CsgTreeNode * const & 
CsgTreeNode::left() const 
{
	return _left;
}

CsgTreeNode * const & 
CsgTreeNode::right() const
{ 
	return _right;
}

CsgTreeNode *
CsgTreeNode::normalised()
{
	if (type()==PRIMITIVE_NODE)
		return this;

	CsgTreeNode *node = this;

	for (;;)
	{
		int finished = 0;

		//
		// Match production rules 1-6
		//

		while (!finished)
		{
			switch (node->type())
			{
			case UNION_NODE:
			case PRIMITIVE_NODE:
				finished = 1;
				break;

			case DIFFERENCE_NODE:
				switch (node->right()->type())
				{
				case UNION_NODE:
					node = productionRule1(node);
					break;
				case INTERSECTION_NODE:
					node = productionRule3(node);
					break;
				case DIFFERENCE_NODE:
					node = productionRule5(node);
					break;
				default:
					finished = 1;
				}
				break;

			case INTERSECTION_NODE:
				switch (node->right()->type())
				{
				case UNION_NODE:
					node = productionRule2(node);
					break;
				case INTERSECTION_NODE:
					node = productionRule4(node);
					break;
				case DIFFERENCE_NODE:
					node = productionRule6(node);
					break;
				default:
					finished = 1;
				}
				break;
			}

			if (finished)
				if (node->left()->type()==UNION_NODE)
				{
					finished = 0;

					if (node->type()==DIFFERENCE_NODE)
						node = productionRule7(node);
					else
						if (node->type()==INTERSECTION_NODE)
							node = productionRule8(node);
						else
							finished = 1;
				}
		}

		//
		// Normalise left sub-tree
		//

		node->left() = node->left()->normalised();

		if (node->type()==UNION_NODE)
			break;

		if (node->right()->type()==PRIMITIVE_NODE && node->left()->type()!=UNION_NODE)
			break;

	}

	//
	// Normalise right sub-tree
	//

	node->right() = node->right()->normalised();

	return node;
}

//
// Production rule 1
//

CsgTreeNode *
CsgTreeNode::productionRule1(CsgTreeNode *node)
{
	#ifdef _DEBUG
	cout << "Production rule 1" << endl;
	#endif

	CsgTreeNode *newNode1 = new CsgTreeNode(DIFFERENCE_NODE);
	CsgTreeNode *newNode2 = new CsgTreeNode(DIFFERENCE_NODE);

	newNode1->left()  = newNode2;
	newNode1->right() = node->right()->right()->clone();
	newNode2->left()  = node->left()->clone();
	newNode2->right() = node->right()->left()->clone();

	return newNode1;
}

//
// Production rule 2
//

CsgTreeNode *
CsgTreeNode::productionRule2(CsgTreeNode *node)
{
	#ifdef _DEBUG
	cout << "Production rule 2" << endl;
	#endif

	CsgTreeNode *newNode1 = new CsgTreeNode(UNION_NODE);
	CsgTreeNode *newNode2 = new CsgTreeNode(INTERSECTION_NODE);
	CsgTreeNode *newNode3 = new CsgTreeNode(INTERSECTION_NODE);

	newNode1->left()  = newNode2;
	newNode1->right() = newNode3;
	newNode2->left()  = node->left()->clone();
	newNode2->right() = node->right()->left()->clone();
	newNode3->left()  = node->left()->clone();
	newNode3->right() = node->right()->right()->clone();

	return newNode1;
}

//
// Production rule 3
//

CsgTreeNode *
CsgTreeNode::productionRule3(CsgTreeNode *node)
{
	#ifdef _DEBUG
	cout << "Production rule 3" << endl;
	#endif

	CsgTreeNode *newNode1 = new CsgTreeNode(UNION_NODE);
	CsgTreeNode *newNode2 = new CsgTreeNode(DIFFERENCE_NODE);
	CsgTreeNode *newNode3 = new CsgTreeNode(DIFFERENCE_NODE);

	newNode1->left()  = newNode2;
	newNode1->right() = newNode3;
	newNode2->left()  = node->left()->clone();
	newNode2->right() = node->right()->left()->clone();
	newNode3->left()  = node->left()->clone();
	newNode3->right() = node->right()->right()->clone();

	return newNode1;
}

//
// Production rule 4
//

CsgTreeNode *
CsgTreeNode::productionRule4(CsgTreeNode *node)
{
	#ifdef _DEBUG
	cout << "Production rule 4" << endl;
	#endif

	CsgTreeNode *newNode1 = new CsgTreeNode(INTERSECTION_NODE);
	CsgTreeNode *newNode2 = new CsgTreeNode(INTERSECTION_NODE);

	newNode1->left()  = newNode2;
	newNode1->right() = node->right()->right()->clone();
	newNode2->left()  = node->left()->clone();
	newNode2->right() = node->right()->left()->clone();

	return newNode1;
}

//
// Production rule 5
//

CsgTreeNode *
CsgTreeNode::productionRule5(CsgTreeNode *node)
{
	#ifdef _DEBUG
	cout << "Production rule 5" << endl;
	#endif

	CsgTreeNode *newNode1 = new CsgTreeNode(UNION_NODE);
	CsgTreeNode *newNode2 = new CsgTreeNode(DIFFERENCE_NODE);
	CsgTreeNode *newNode3 = new CsgTreeNode(INTERSECTION_NODE);

	newNode1->left()  = newNode2;
	newNode1->right() = newNode3;
	newNode2->left()  = node->left()->clone();
	newNode2->right() = node->right()->left()->clone();
	newNode3->left()  = node->left()->clone();
	newNode3->right() = node->right()->right()->clone();

	return newNode1;
}

//
// Production rule 6
//

CsgTreeNode *
CsgTreeNode::productionRule6(CsgTreeNode *node)
{
	#ifdef _DEBUG
	cout << "Production rule 6" << endl;
	#endif

	CsgTreeNode *newNode1 = new CsgTreeNode(DIFFERENCE_NODE);
	CsgTreeNode *newNode2 = new CsgTreeNode(INTERSECTION_NODE);

	newNode1->left()  = newNode2;
	newNode1->right() = node->right()->right()->clone();
	newNode2->left()  = node->left()->clone();
	newNode2->right() = node->right()->left()->clone();

	return newNode1;
}

//
// Production rule 7
//

CsgTreeNode *
CsgTreeNode::productionRule7(CsgTreeNode *node)
{
	#ifdef _DEBUG
	cout << "Production rule 7" << endl;
	#endif

	CsgTreeNode *newNode1 = new CsgTreeNode(UNION_NODE);
	CsgTreeNode *newNode2 = new CsgTreeNode(DIFFERENCE_NODE);
	CsgTreeNode *newNode3 = new CsgTreeNode(DIFFERENCE_NODE);

	newNode1->left()  = newNode2;
	newNode1->right() = newNode3;
	newNode2->left()  = node->left()->left()->clone();
	newNode2->right() = node->right()->clone();
	newNode3->left()  = node->left()->right()->clone();
	newNode3->right() = node->right()->clone();

	return newNode1;
}

//
// Production rule 8
//

CsgTreeNode *
CsgTreeNode::productionRule8(CsgTreeNode *node)
{
	#ifdef _DEBUG
	cout << "Production rule 8" << endl;
	#endif

	CsgTreeNode *newNode1 = new CsgTreeNode(UNION_NODE);
	CsgTreeNode *newNode2 = new CsgTreeNode(INTERSECTION_NODE);
	CsgTreeNode *newNode3 = new CsgTreeNode(INTERSECTION_NODE);

	newNode1->left()  = newNode2;
	newNode1->right() = newNode3;
	newNode2->left()  = node->left()->left()->clone();
	newNode2->right() = node->right()->clone();
	newNode3->left()  = node->left()->right()->clone();
	newNode3->right() = node->right()->clone();

	return newNode1;
}

/*
GLubyte 
CsgTreeNode::countSurfaces(const vector<CsgTreeLeaf *> &leaf,const countSurfacesMode mode)
{
	GLuint dl = glGenLists(1);
	glNewList(dl,GL_COMPILE);
	for (int p=0; p<leaf.size(); p++)
		leaf[p]->draw();
	glEndList();

	GLubyte count = ::countSurfaces(&dl,1,mode);
	glDeleteLists(dl,1);

	return count;
}
*/

void 
CsgTreeNode::draw()
{
	if (type()==PRIMITIVE_NODE)
		draw();
	else
	{
		left()->draw();
		right()->draw();
	}
}

void
CsgTreeNode::normalise()
{
	startAllocList();
		CsgTreeNode *normalisedTree = normalised();
		CsgSumOfProducts sum = CsgSumOfProducts();
		_sumOfProducts = sum;
		normalisedTree->buildProducts(_sumOfProducts);
	endAllocList();
	deleteAllocList();
}

CsgProduct  
CsgTreeNode::optimiseProduct(const CsgProduct &product)
{
	CsgProduct tmp;

	// If a product contains no uncomplemented primitives 
	// then it can be treated as empty.

	if (product.uncomplemented().size()==0) return tmp;

	// If any uncomplmented primitive matches a complemented
	// product then the product is empty.

	for (unsigned int u=0; u<product.uncomplemented().size(); u++)
		for (unsigned int c=0; c<product.complemented().size(); c++)
			if (!strcmp(product.uncomplemented()[u]->name(),product.complemented()[c]->name()))
				return tmp;

	// To-do:  Remove duplicates

	return product;
}

void 
CsgTreeNode::buildProducts(CsgSumOfProducts &sumOfProducts)
{
	if (type()==UNION_NODE)
	{
		left()->buildProducts(sumOfProducts);
		right()->buildProducts(sumOfProducts);
	}
	else
	{
		CsgProduct prod;
		buildProduct(prod,false);

		CsgProduct optProd = optimiseProduct(prod);
		if (optProd.size()>0)
			sumOfProducts.push_back(optProd);
	}
}

void 
CsgTreeNode::buildProduct(CsgProduct &product,const bool invert)
{
	switch (type())
	{
	case PRIMITIVE_NODE:
		if (invert)
			product.complemented().push_back((CsgTreeLeaf *) this);
		else
			product.uncomplemented().push_back((CsgTreeLeaf *) this);
		break;

	case UNION_NODE:
		assert(0);
		break;

	case INTERSECTION_NODE:
		assert(!invert);
		left()->buildProduct(product,invert);
		right()->buildProduct(product,invert);
		break;

	case DIFFERENCE_NODE:
		assert(!invert);
		left()->buildProduct(product,invert);
		right()->buildProduct(product,true);
		break;
	}
}

ostream &
operator<<(ostream &os,CsgTreeNode &node)
{
	if (node.type()==PRIMITIVE_NODE)
		os << ((CsgTreeLeaf *) &node)->name();
	else
	{
		if (node.left()->type()!=PRIMITIVE_NODE)
			os << "( " << *node.left() << " )";
		else
			os << *node.left() ;

		switch (node.type())
		{
			case UNION_NODE:		os << " or "; break;
			case INTERSECTION_NODE: os << " and "; break;
			case DIFFERENCE_NODE:	os << " - "; break;
		}

		if (node.right()->type()!=PRIMITIVE_NODE)
			os << "( " << *node.right() << " )";
		else
			os << *node.right() ;
	}

	return os;
}

CsgTreeLeaf::CsgTreeLeaf
(
	char *name,
	const GLuint displayList,
	GLfloat *colour,
	const Matrix position
)
: 
  CsgTreeNode(PRIMITIVE_NODE), 
  Matrix(position), 
  _name(NULL), 
  _colour(colour), 
  _displayList(displayList),
  _convexity(1)
{
	  setName(name);
}

CsgTreeLeaf::CsgTreeLeaf(const CsgTreeLeaf &node)
: 
  CsgTreeNode(node), 
  Matrix(node), 
  _name(NULL), 
  _colour(node._colour), 
  _displayList(node._displayList),
  _convexity(1)
{
	  setName(node._name);
}

CsgTreeLeaf::~CsgTreeLeaf()
{
	#ifdef _DEBUG
	cout << "Calling delete for CsgTreePrimitive (" << _name << ") " << (void *) this << endl;
	#endif

	if (_name)
		delete _name;
}

char *
CsgTreeLeaf::name() const
{
	return _name;
}

void
CsgTreeLeaf::setName(char *name)
{
	if (_name)
		delete _name;

	_name = new char [ strlen(name) + 1 ];
	strcpy(_name,name);
}

void 
CsgTreeLeaf::draw()
{
	glPushMatrix();
		if (_colour)
			glColor3fv(_colour);
		glMultMatrix();
		glCallList(_displayList);
	glPopMatrix();
}

void 
CsgTreeLeaf::drawFront()
{
	glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT);

		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);
		draw();

	glPopAttrib();
}

void 
CsgTreeLeaf::drawBack()
{
	glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT);

		glEnable(GL_CULL_FACE);
		glCullFace(GL_FRONT);
		draw();

	glPopAttrib();
}

Uint &
CsgTreeLeaf::convexity() 
{
	return _convexity;
}

//
// Intersect
//
// Result in stencil should be one for parity test to pass.
//

void 
CsgTreeLeaf::intersect(const int bit,GLuint &testMask,const GLenum depthFunc)
{
	parityTest(bit,depthFunc);
	testMask |= (1<<bit);
}

//
// Subtract
//
// Result in stencil should be zero for parity test to pass.
//

void 
CsgTreeLeaf::subtract(const int bit,GLuint &testMask,const GLenum depthFunc)
{
	parityTest(bit,depthFunc);
}

//
// Apply parity test to frame buffer
//

void 
CsgTreeLeaf::parityTest(const int bit,const GLenum depthFunc)
{
	GLuint mask = (1<<bit);

	glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

		glEnable(GL_STENCIL_TEST);
		glStencilFunc(GL_ALWAYS,0xffff,0xffff);
		glStencilOp(GL_KEEP,GL_KEEP,GL_INVERT);
		glStencilMask(mask);

		glEnable(GL_DEPTH_TEST);
		glDepthFunc(depthFunc);
		glDisable(GL_CULL_FACE);
		glDepthMask(GL_FALSE);
		glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);

		draw();

	glPopAttrib();
}


