//
// OpenGL Solid Modelling
//
// Nigel Stewart (nigels@eisa.net.au)
// RMIT Department of Mechanical and Manufacturing Engineering
// 11th December 1997
//

#ifndef CSGTREE_H
#define CSGTREE_H

#include <vector.h>
#include <list.h>

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glut.h>

#include <iostream.h>

#include <vector3.h>
#include <matrix4.h>

#include <glbuffer.h>
#include <cntsurf.h>

//////////////////////////////////////////////////
// CSG Product
//
// A CSG product consists of a series of primitive
// objects which are related by intersection or
// boolean difference.  We model them in this
// implementation by maintaining two vectors: one
// for uncomplemented primitves, the other for 
// subtracted (or complemented) primitives.
//
// For example:  A.B-C-D
//
// uncomplemented = { A , B }
// complemented   = { C , D }
//
// Note that there should always be at least one
// uncomplemented primitve.

class CsgTreeNode;
class CsgTreeLeaf;

class CsgProduct
{
public:

	CsgProduct();
	CsgProduct(const CsgProduct &prod);
	CsgProduct &operator=(const CsgProduct &prod);
	
	vector<CsgTreeLeaf *> &uncomplemented();
	vector<CsgTreeLeaf *> &complemented();

	const vector<CsgTreeLeaf *> &uncomplemented() const;
	const vector<CsgTreeLeaf *> &complemented() const;

	int size() const;

	friend ostream &operator<<(ostream &os,CsgProduct &prod);

	// Access as one array

	CsgTreeLeaf *element(const Uint i) const;
	bool elementComplemented(const Uint i) const;

private:

	vector<CsgTreeLeaf *> _uncomplemented;
	vector<CsgTreeLeaf *> _complemented;
};

//////////////////////////////////////////////////
// CSG Sum of Products
//
// Any CSG expression can be represented as a 
// sum of products.  The Goldfeather article
// describes one method of converting from an
// arbitrary CSG expression to sum of products
// form.
//
// We model a sum of products by using a vector
// of CSG products.
//

class CsgSumOfProducts : public vector<CsgProduct>
{
public:
	CsgSumOfProducts();
	CsgSumOfProducts &operator=(const CsgSumOfProducts &sum);
	
	friend ostream &operator<<(ostream &os,CsgSumOfProducts &sumOfProd);
};

//////////////////////////////////////////////////
// CSG Tree Node
//
// A CSG tree is formed by a collection of leaf and
// boolean operation nodes.  Primitive objects are
// stored at each leaf and are combined via the
// boolean operators at their parent nodes.  
//
// We enumerate the type of the csg tree node 
// explicitly.  The implementation for primitives
// has been seperated in order to support alternative
// mechanisms to the OpenGL display list. 
//

typedef enum
{
	PRIMITIVE_NODE,
	UNION_NODE,
	INTERSECTION_NODE,
	DIFFERENCE_NODE
} CsgTreeNodeType;

class CsgPrimitiveDatabase;

class CsgTreeNode
{
public:

	CsgTreeNode(const CsgTreeNodeType  type);	
	CsgTreeNode(const CsgTreeNode     &type);	// Copy constructor

	virtual ~CsgTreeNode();

	CsgTreeNodeType type() const;

	CsgTreeNode *&left();
	CsgTreeNode *&right();
	CsgTreeNode *const& left() const;
	CsgTreeNode *const& right() const;

	virtual void draw();

	friend ostream &operator<<(ostream &os,CsgTreeNode &node);

	//
	// Image-space CSG rendering functions
	//

	void normalise();
	CsgSumOfProducts &sumOfProducts() { return _sumOfProducts; }

private:

	CsgSumOfProducts _sumOfProducts;

	// Binary tree node info

	CsgTreeNodeType      _nodeType;
	CsgTreeNode         *_left,*_right;

	// Here we implement tree normalisation functionality

	CsgTreeNode *clone();

	CsgTreeNode *normalised();

	CsgTreeNode *productionRule1(CsgTreeNode *node);
	CsgTreeNode *productionRule2(CsgTreeNode *node);
	CsgTreeNode *productionRule3(CsgTreeNode *node);
	CsgTreeNode *productionRule4(CsgTreeNode *node);
	CsgTreeNode *productionRule5(CsgTreeNode *node);
	CsgTreeNode *productionRule6(CsgTreeNode *node);
	CsgTreeNode *productionRule7(CsgTreeNode *node);
	CsgTreeNode *productionRule8(CsgTreeNode *node);

	void		buildProducts  (CsgSumOfProducts &sumOfProducts );
	void		buildProduct   (CsgProduct &product,const bool invert);
	CsgProduct  optimiseProduct(const CsgProduct &product);

	// Boolean rendering functions

//	GLubyte countSurfaces(const vector<CsgTreeLeaf *> &leaf,const countSurfacesMode mode);

//	void readDepthBuffer();
//	void writeDepthBuffer();

	// Tree node memory management

	static list<CsgTreeNode *>   _allocList;
	static bool                  _allocListEnable;

	void startAllocList();
	void endAllocList();
	void deleteAllocList();
};

////////////////
// CSG Tree leaf

class CsgTreeLeaf : public CsgTreeNode, public Matrix
{
	friend CsgPrimitiveDatabase;
	friend CsgTreeNode;
	friend void GoldfeatherRenderingAlgorithm(const CsgSumOfProducts &sum,bool copy);
	friend void LayeredGoldfeatherRenderingAlgorithm(const CsgSumOfProducts &sum,bool copy);

public:

	// Contruction/Destruction

	CsgTreeLeaf
	(
		char *name,
		const GLuint displayList,
		GLfloat *colour = NULL,
		const Matrix position = Matrix()
	);

	CsgTreeLeaf(const CsgTreeLeaf &leaf);
	
	virtual ~CsgTreeLeaf();

	// Configuration

	char *name() const;
	void  setName(char *name);

	GLfloat *color() const { return _colour; }

	Uint &convexity();

	// Rendering

	virtual void drawFront();		// Draw front-facing sufraces into frame buffer
	virtual void drawBack();		// Draw back-facing surfaces into frame buffer

	virtual void draw();

private:

	// Count number of surfaces in front of z
	void parityTest(const int bit,const GLenum depthFunc = GL_LESS);	

	virtual void intersect(const int bit,GLuint &testMask,const GLenum depthFunc = GL_LESS);	
	virtual void subtract(const int bit,GLuint &testMask,const GLenum depthFunc = GL_LESS);

	// Primitive info

	char *_name;
	GLfloat *_colour;
	GLuint _displayList;
	Uint _convexity;
};

#endif
