#include <glbuffer.h>

#include <string.h>
#include <assert.h>
#include <stdio.h>

OpenGLbuffer::OpenGLbuffer()
{
	GLuint viewport[4];
	glGetIntegerv(GL_VIEWPORT,(GLint *) viewport);

	_x = viewport[0]; 
	_y = viewport[1];
	_width = viewport[2];
	_height = viewport[3];
	_size = _width*_height;
}

GLuint OpenGLbuffer::x() const		{ return _x; }
GLuint OpenGLbuffer::y() const		{ return _y; }
GLuint OpenGLbuffer::width() const	{ return _width; }
GLuint OpenGLbuffer::height() const	{ return _height; }
GLuint OpenGLbuffer::size() const	{ return _size; }

OpenGLcolourBuffer::OpenGLcolourBuffer()
: OpenGLbuffer()
{
	_pixels = new GLubyte[_size*3];

	read();
}

OpenGLcolourBuffer::OpenGLcolourBuffer(GLubyte *pixels)
: OpenGLbuffer(), _pixels(pixels)
{
}

OpenGLcolourBuffer::~OpenGLcolourBuffer()
{
	delete [] _pixels;
}

void
OpenGLcolourBuffer::read()
{
	//
	// Check that viewport has not changed since
	// construction.
	//

	GLuint viewport[4];
	glGetIntegerv(GL_VIEWPORT,(GLint *) viewport);

	if (_width!=viewport[2] || _height!=viewport[3] || _pixels==NULL)
	{
		_x = viewport[0];
		_y = viewport[1];
		_width = viewport[2];
		_height = viewport[3];
		_size = _width*_height;

		if (_pixels)
			delete[] _pixels;
		_pixels = new GLubyte[_size*3];
	}

	//
	// Read the depth buffer
	//

	glPixelStorei(GL_PACK_ALIGNMENT,sizeof(GLubyte));
	glReadPixels(_x,_y,_width,_height,GL_RGB,GL_UNSIGNED_BYTE,_pixels);
}

void 
OpenGLcolourBuffer::write() const
{
	glPixelStorei(GL_UNPACK_ALIGNMENT,sizeof(GLubyte));
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(-1.0,1.0,-1.0,1.0,0.0,100.0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glRasterPos2f(-1.0F,-1.0F);
	glDrawPixels(_width,_height,GL_RGB,GL_UNSIGNED_BYTE,_pixels);
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	
	glMatrixMode(GL_MODELVIEW);
}

void 
OpenGLcolourBuffer::writePPM(char *filename) const
{
	FILE *fout;
	
	fout = fopen(filename,"wb");

	unsigned char buffer[1024];
	sprintf((char *) buffer,"P6 %u %u 255\n",_width,_height);

	fwrite(buffer,1,strlen((char *) buffer),fout);

	for (GLuint y=_height-1;y<_height;--y) 
		fwrite(_pixels+y*_width*3,1,_width*3,fout);

	fclose(fout);
}

void 
OpenGLcolourBuffer::writeTGA(char *filename) const
{
	GLubyte tgaHeader[18];

	tgaHeader[0] = 0;
	tgaHeader[1] = 0;
	tgaHeader[2] = 2;
	tgaHeader[3] = 0;
	tgaHeader[4] = 0;
	tgaHeader[5] = 0;
	tgaHeader[6] = 0;
	tgaHeader[7] = 0;
	tgaHeader[8] = 0;
	tgaHeader[9] = 0;
	tgaHeader[10] = 0;
	tgaHeader[11] = 0;
	tgaHeader[12] = width()&0xff;
	tgaHeader[13] = (width()>>8)&0xff;
	tgaHeader[14] = height()&0xff;
	tgaHeader[15] = (height()>>8)&0xff;
	tgaHeader[16] = 24;
	tgaHeader[17] = 0;

	FILE *fout;

	fout  = fopen(filename,"wb");
	fwrite(tgaHeader,1,18,fout);

	GLubyte rgb[3];

	for ( unsigned int y=0; y<height(); y++ )
		for ( unsigned int x=0; x<width(); x++ )
		{
			rgb[0] = _pixels[(x+y*width())*3+2];
			rgb[1] = _pixels[(x+y*width())*3+1];
			rgb[2] = _pixels[(x+y*width())*3];

			fwrite(rgb,1,3,fout);
		}

	fclose(fout);

//-------------------------------------------------------------------------------
//DATA TYPE 2:  Unmapped RGB images.                                             |
//_______________________________________________________________________________|
//| Offset | Length |                     Description                            |
//|--------|--------|------------------------------------------------------------|
//|--------|--------|------------------------------------------------------------|
//|    0   |     1  |  Number of Characters in Identification Field.             |
//|        |        |                                                            |
//|        |        |  This field is a one-byte unsigned integer, specifying     |
//|        |        |  the length of the Image Identification Field.  Its value  |
//|        |        |  is 0 to 255.  A value of 0 means that no Image            |
//|        |        |  Identification Field is included.                         |
//|        |        |                                                            |
//|--------|--------|------------------------------------------------------------|
//|    1   |     1  |  Color Map Type.                                           |
//|        |        |                                                            |
//|        |        |  This field contains either 0 or 1.  0 means no color map  |
//|        |        |  is included.  1 means a color map is included, but since  |
//|        |        |  this is an unmapped image it is usually ignored.  TIPS    |
//|        |        |  ( a Targa paint system ) will set the border color        |
//|        |        |  the first map color if it is present.                     |
//|        |        |                                                            |
//|--------|--------|------------------------------------------------------------|
//|    2   |     1  |  Image Type Code.                                          |
//|        |        |                                                            |
//|        |        |  This field will always contain a binary 2.                |
//|        |        |  ( That's what makes it Data Type 2 ).                     |
//|        |        |                                                            |
//|--------|--------|------------------------------------------------------------|
//|    3   |     5  |  Color Map Specification.                                  |
//|        |        |                                                            |
//|        |        |  Ignored if Color Map Type is 0; otherwise, interpreted    |
//|        |        |  as follows:                                               |
//|        |        |                                                            |
//|    3   |     2  |  Color Map Origin.                                         |
//|        |        |  Integer ( lo-hi ) index of first color map entry.         |
//|        |        |                                                            |
//|    5   |     2  |  Color Map Length.                                         |
//|        |        |  Integer ( lo-hi ) count of color map entries.             |
//|        |        |                                                            |
//|    7   |     1  |  Color Map Entry Size.                                     |
//|        |        |  Number of bits in color map entry.  16 for the Targa 16,  |
//|        |        |  24 for the Targa 24, 32 for the Targa 32.                 |
//|        |        |                                                            |
//|--------|--------|------------------------------------------------------------|
//|    8   |    10  |  Image Specification.                                      |
//|        |        |                                                            |
//|    8   |     2  |  X Origin of Image.                                        |
//|        |        |  Integer ( lo-hi ) X coordinate of the lower left corner   |
//|        |        |  of the image.                                             |
//|        |        |                                                            |
//|   10   |     2  |  Y Origin of Image.                                        |
//|        |        |  Integer ( lo-hi ) Y coordinate of the lower left corner   |
//|        |        |  of the image.                                             |
//|        |        |                                                            |
//|   12   |     2  |  Width of Image.                                           |
//|        |        |  Integer ( lo-hi ) width of the image in pixels.           |
//|        |        |                                                            |
//|   14   |     2  |  Height of Image.                                          |
//|        |        |  Integer ( lo-hi ) height of the image in pixels.          |
//|        |        |                                                            |
//|   16   |     1  |  Image Pixel Size.                                         |
//|        |        |  Number of bits in a pixel.  This is 16 for Targa 16,      |
//|        |        |  24 for Targa 24, and .... well, you get the idea.         |
//|        |        |                                                            |
//|   17   |     1  |  Image Descriptor Byte.                                    |
//|        |        |  Bits 3-0 - number of attribute bits associated with each  |
//|        |        |             pixel.  For the Targa 16, this would be 0 or   |
//|        |        |             1.  For the Targa 24, it should be 0.  For     |
//|        |        |             Targa 32, it should be 8.                      |
//|        |        |  Bit 4    - reserved.  Must be set to 0.                   |
//|        |        |  Bit 5    - screen origin bit.                             |
//|        |        |             0 = Origin in lower left-hand corner.          |
//|        |        |             1 = Origin in upper left-hand corner.          |
//|        |        |             Must be 0 for Truevision images.               |
//|        |        |  Bits 7-6 - Data storage interleaving flag.                |
//|        |        |             00 = non-interleaved.                          |
//|        |        |             01 = two-way (even/odd) interleaving.          |
//|        |        |             10 = four way interleaving.                    |
//|        |        |             11 = reserved.                                 |
//|        |        |                                                            |
//|--------|--------|------------------------------------------------------------|
//|   18   | varies |  Image Identification Field.                               |
//|        |        |  Contains a free-form identification field of the length   |
//|        |        |  specified in byte 1 of the image record.  It's usually    |
//|        |        |  omitted ( length in byte 1 = 0 ), but can be up to 255    |
//|        |        |  characters.  If more identification information is        |
//|        |        |  required, it can be stored after the image data.          |
//|        |        |                                                            |
//|--------|--------|------------------------------------------------------------|
//| varies | varies |  Color map data.                                           |
//|        |        |                                                            |
//|        |        |  If the Color Map Type is 0, this field doesn't exist.     |
//|        |        |  Otherwise, just read past it to get to the image.         |
//|        |        |  The Color Map Specification describes the size of each    |
//|        |        |  entry, and the number of entries you'll have to skip.     |
//|        |        |  Each color map entry is 2, 3, or 4 bytes.                 |
//|        |        |                                                            |
//|--------|--------|------------------------------------------------------------|
//| varies | varies |  Image Data Field.                                         |
//|        |        |                                                            |
//|        |        |  This field specifies (width) x (height) pixels.  Each     |
//|        |        |  pixel specifies an RGB color value, which is stored as    |
//|        |        |  an integral number of bytes.                              |
//|        |        |                                                            |
//|        |        |  The 2 byte entry is broken down as follows:               |
//|        |        |  ARRRRRGG GGGBBBBB, where each letter represents a bit.    |
//|        |        |  But, because of the lo-hi storage order, the first byte   |
//|        |        |  coming from the file will actually be GGGBBBBB, and the   |
//|        |        |  second will be ARRRRRGG. "A" represents an attribute bit. |
//|        |        |                                                            |
//|        |        |  The 3 byte entry contains 1 byte each of blue, green,     |
//|        |        |  and red.                                                  |
//|        |        |                                                            |
//|        |        |  The 4 byte entry contains 1 byte each of blue, green,     |
//|        |        |  red, and attribute.  For faster speed (because of the     |
//|        |        |  hardware of the Targa board itself), Targa 24 images are  |
//|        |        |  sometimes stored as Targa 32 images.                      |
//|        |        |                                                            |
//--------------------------------------------------------------------------------
}


