//
// OpenGL Solid Modelling
//
// Nigel Stewart (nigels@eisa.net.au)
// RMIT Department of Mechanical and Manufacturing Engineering
// 11th December 1997
//

#ifndef GLBUFFER_H
#define GLBUFFER_H

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#include <assert.h>
#include <iostream.h>
#include <stdio.h>
#include <string.h>

#define GLERROR \
{ \
  GLenum code = glGetError(); \
  while (code!=GL_NO_ERROR) \
  { \
    cerr << __FILE__ << '(' << __LINE__ << ") " << (char *) gluErrorString(code) << endl; \
    code = glGetError();	\
  } 	\
}

/////////////////////////// OpenGLbuffer ////////////////////////////

//
// OpenGLbuffer is a base class for colour, depth and
// stencil buffer classes.  These share common functionality
// such as width, height and functions to write data to disk.
//

class OpenGLbuffer
{
public:

	virtual void read() = 0;		    // Update from frame buffer
	virtual void write() const = 0;		// Write to frame buffer
	
	virtual void writePPM(char *filename) const = 0;

	GLuint x() const;		    // Viewport x
	GLuint y() const;		    // Viewport y
	GLuint width() const;		// Viewport width
	GLuint height() const;		// Viewport height
	GLuint size() const;		// Viewport size in pixels

protected:

	OpenGLbuffer();

	GLuint _x,_y;
	GLuint _width,_height;
	GLuint _size;
};

/////////////////////////// OpenGLcolourBuffer ////////////////////////////

class OpenGLcolourBuffer : public OpenGLbuffer
{
public:

	OpenGLcolourBuffer();
	OpenGLcolourBuffer(GLubyte *pixels);
	virtual ~OpenGLcolourBuffer();

	void read();
	void write() const;
	void writePPM(char *filename) const;
	void writeTGA(char *filename) const;

protected:

	GLubyte *_pixels;
};

// GLfloat,  GL_FLOAT
// GLushort, GL_UNSIGNED_SHORT
// GLuint,   GL_UNSIGNED_INT

#ifdef TEMPLATE_BUFFERS
template< class DepthType , GLenum GlDepthType >
#else
#define DepthType GLushort
#define GlDepthType GL_UNSIGNED_INT
#endif
class OpenGLdepthBuffer : public OpenGLbuffer
{
protected:

	DepthType *_pixels;

public:

/*************************************************

//
// Public Interface
//

	OpenGLdepthBuffer();
	OpenGLdepthBuffer(DepthType *pixels);
	~OpenGLdepthBuffer();

	OpenGLdepthBuffer &operator=(const DepthType &value);

	void read();
	void write() const;
	void writePPM(char *filename) const;

	DepthType maxPossible() const;
	DepthType minPossible() const;
	
***************************************************/

// Implementation
	

	OpenGLdepthBuffer() : OpenGLbuffer()
	{
		_pixels = new DepthType[_size];
	        read();
	};

	OpenGLdepthBuffer(DepthType *pixels) : OpenGLbuffer(), _pixels(pixels)
	{
	};
		                
	~OpenGLdepthBuffer()
	{
		delete [] _pixels;
	};
	
	OpenGLdepthBuffer & 
	operator=(const DepthType &value)
	{
	    // Check that viewport has not changed since
	    // construction.
	    	
	    GLuint viewport[4];
	    glGetIntegerv(GL_VIEWPORT,(GLint *) viewport);
	    	
		if (_width!=viewport[2] || _height!=viewport[3] || _pixels==NULL)
		{
			_x = viewport[0];
			_y = viewport[1];
			_width = viewport[2];
			_height = viewport[3];
			_size = _width*_height;
			
			if (_pixels)
				delete [] _pixels;
				
			_pixels = new DepthType[_size];
		}

		for (unsigned int n=0; n<_size; n++)
			_pixels[n] = value;

		return *this;
	}


	void 
	read()
	{	
	    // Check that viewport has not changed since
	    // construction.
	    	
	    GLuint viewport[4];
	    glGetIntegerv(GL_VIEWPORT,(GLint *) viewport);
	    	
		if (_width!=viewport[2] || _height!=viewport[3] || _pixels==NULL)
		{
			_x = viewport[0];
			_y = viewport[1];
			_width = viewport[2];
			_height = viewport[3];
			_size = _width*_height;
			
			if (_pixels)
				delete [] _pixels;
				
			_pixels = new DepthType[_size];
		}
		
		glPixelStorei(GL_PACK_ALIGNMENT,sizeof(DepthType));
		glReadPixels(_x,_y,_width,_height,GL_DEPTH_COMPONENT,GlDepthType,_pixels);
	};
	
	void
	write() const
	{
		glPixelStorei(GL_UNPACK_ALIGNMENT,sizeof(DepthType));
		
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		
		glLoadIdentity();
		glOrtho(-1.0,1.0,-1.0,1.0,0.0,100.0);

		glMatrixMode(GL_MODELVIEW);
	        glPushMatrix();
	        glLoadIdentity();

		glRasterPos2f(-1.0F,-1.0F);
		glDrawPixels(_width,_height,GL_DEPTH_COMPONENT,GlDepthType,_pixels);
		
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
	};
	
	void 
	writePPM(char *filename) const
	{
		FILE *fout;	
		fout = fopen(filename,"wb");

		unsigned char buffer[1024];
		sprintf((char *) buffer,"P5 %u %u 255\n",_width,_height);

		fwrite(buffer,1,strlen((char *) buffer),fout);

		DepthType max = maxPossible();
		DepthType range = maxPossible()-minPossible();

		for (GLuint y=_height-1;y<_height;--y) 
			for (GLuint x=0;x<_width;++x) 
			{
				DepthType depth = _pixels[x+y*_width];

				float cf = ((float) max - (float) depth)/((float) range);
				unsigned char color = (unsigned char) ((float) cf*255.0);

				buffer[0] = color;

				fwrite(buffer,1,1,fout);
			}
			
		fclose(fout);
	};

	DepthType
	maxPossible() const
	{
		switch (GlDepthType)
		{
		case GL_UNSIGNED_SHORT:
			return DepthType(0xffff);
		case GL_UNSIGNED_INT:
			#ifdef WIN32
			return DepthType(0xffffffff - 6911);
			#else
			return DepthType(0xffffffff);
			#endif
		case GL_FLOAT:
			return DepthType(1.0);
		default:
			assert(0);
			return DepthType();
		}
	};
	
	DepthType
	minPossible() const
	{
		switch (GlDepthType)
		{
		case GL_UNSIGNED_SHORT:
			return 0x0;
		case GL_UNSIGNED_INT:
			return 0x0;			
		case GL_FLOAT:
			return (DepthType) 0.0;
		default:
			assert(0);
			return DepthType();
		}
	};
};	

// GLubyte,  GL_UNSIGNED_BYTE

#ifdef TEMPLATE_BUFFERS
template<class StencilType,GLenum GlStencilType>
#else
#define StencilType GLubyte
#define GlStencilType GL_UNSIGNED_BYTE
#endif
class OpenGLstencilBuffer : public OpenGLbuffer
{
protected:
	StencilType *_pixels;
	
public:

/****************************************************

//
// Public Interface
//

	OpenGLstencilBuffer();
	OpenGLstencilBuffer(StencilType *pixels);
	~OpenGLstencilBuffer();

	void read();
	void write() const;
	void writePPM(char *filename) const;

*****************************************************/

	OpenGLstencilBuffer() : OpenGLbuffer()
	{
		_pixels = new StencilType[_size];
		read();
	};
	
	OpenGLstencilBuffer(StencilType *pixels)
	: OpenGLbuffer(), _pixels(pixels)
	{
	};
	
	~OpenGLstencilBuffer()
	{
		delete [] _pixels;
	};

	void
	read()
	{
		GLuint viewport[4];
		glGetIntegerv(GL_VIEWPORT,(GLint *) viewport);

		if (_width!=viewport[2] || _height!=viewport[3] || _pixels==NULL)
		{
			_x = viewport[0];
			_y = viewport[1];
			_width = viewport[2];
			_height = viewport[3];
			_size = _width*_height;

			if (_pixels)
				delete[] _pixels;
			_pixels = new StencilType[_size];
		}

		glPixelStorei(GL_PACK_ALIGNMENT,sizeof(StencilType));
		glReadPixels(_x,_y,_width,_height,GL_STENCIL_INDEX,GlStencilType,_pixels);
	};

	void 
	write() const
	{
		glPixelStorei(GL_UNPACK_ALIGNMENT,sizeof(StencilType));
	
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(-1.0,1.0,-1.0,1.0,0.0,100.0);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glRasterPos2f(-1.0F,-1.0F);
		glStencilMask(0xff);
		glDrawPixels(_width,_height,GL_STENCIL_INDEX,GlStencilType,_pixels);
		glPopMatrix();

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
	
		glMatrixMode(GL_MODELVIEW);
	};

	void 
	writePPM(char *filename) const
	{
		FILE *fout;
		fout = fopen(filename,"wb");

		unsigned char buffer[1024];
		sprintf((char *) buffer,"P5 %u %u 255\n",_width,_height);

		fwrite(buffer,1,strlen((char *) buffer),fout);

		for (GLuint y=_height-1;y<_height;--y) 
			for (GLuint x=0;x<_width;++x) 
			{
				GLubyte stencil = _pixels[x+y*_width];
			
				buffer[0] = stencil;
				fwrite(buffer,1,1,fout);
			}

		fclose(fout);
	};
};

#endif
