//
// OpenGL CSG Rendering
//
// Nigel Stewart (nigels@eisa.net.au)
// RMIT Department of Mechanical and Manufacturing Engineering
// 27th December 1998
//
// GoldfeatherRenderingAlgorithm
//
// This module is an OpenGL/C++ implementation
// of the following journal article:
//
// "Near Real-Time CSG Rendering Using Tree Normalisation
// "and Geometric Pruning", Jack Goldfeather, et. al.
// IEEE Computer Graphics and Applications, May 1989, pp 20-28
//

#include <goldfeat.h>

#include <zplane.h>

depthBuffer depthCopy(NULL);
int viewport[4];
bool useCopy;

void GoldfeatherRenderingAlgorithm(const CsgSumOfProducts &sum,bool copy)
{
	// A fakey trick that shouldn't be necessary
	//
	// WindowsNT software OpenGL seems to lose
	// precision on transformation matricies when
	// they are pushed/popped.
	//

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glPopMatrix();

	//

	GLint stencilSize = 0;
	glGetIntegerv(GL_STENCIL_BITS,&stencilSize);

	glGetIntegerv(GL_VIEWPORT,(GLint *) viewport);

	useCopy = copy;

	glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT | GL_POLYGON_BIT | GL_VIEWPORT_BIT);

		if (useCopy)
			glViewport(0,0,viewport[2]>>1,viewport[3]);

		bool firstTime = true;

		// Do not draw into RGBA, or do expensive
		// things like lighting or normalisation

		glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
		glDisable(GL_LIGHTING);
		glDisable(GL_NORMALIZE);

		for (int t=0; t<sum.size(); t++)
		{
			// Preserve the current z-buffer if this
			// is not the first pass.

			if (!firstTime)
				readDepthBuffer();

			// For products of size one,
			// no clipping is necessary

			if (sum[t].size()==1)
			{
				glDisable(GL_STENCIL_TEST);
				glEnable(GL_DEPTH_TEST);
				glDepthFunc(GL_LESS);
				glDepthMask(GL_TRUE);

				// Draw front surface with z-less test

				while (t<sum.size() && sum[t].size()==1)
					sum[t++].uncomplemented()[0]->drawFront();

				// Merge with saved z-buffer, if necessary

				if (firstTime)
					firstTime = false;
				else
					writeDepthBuffer();
			}
			else
			{
				// Non-trivial products must be 
				// clipped, one primitive at a time.

				for (int p=0; p<sum[t].size(); p++)
					for (Uint k=0; k<sum[t].element(p)->convexity(); k++)
					{
						#ifdef _DEBUG
						cout << "p = " << p << ", k=" << k << endl;
						#endif

						if (p!=0 || k!=0)
							readDepthBuffer();

						// Draw front or back surface into
						// z-buffer

						if (sum[t].element(p)->convexity()==1)
						{
							glDepthFunc(GL_ALWAYS);
							glDepthMask(GL_TRUE);

							glDisable(GL_STENCIL_TEST);
							glEnable(GL_CULL_FACE);

							if (sum[t].elementComplemented(p))
								sum[t].element(p)->drawBack();
							else
								sum[t].element(p)->drawFront();
						}
						else
						{
							glDepthFunc(GL_ALWAYS);
							glDepthMask(GL_TRUE);

							glEnable(GL_STENCIL_TEST);
							glStencilOp(GL_INCR,GL_INCR,GL_INCR);
							glStencilFunc(GL_EQUAL,k,0xffff);
							glStencilMask(0xffff);

							glEnable(GL_CULL_FACE);
							if (sum[t].elementComplemented(p))
								sum[t].element(p)->drawBack();
							else
								sum[t].element(p)->drawFront();

							glClear(GL_STENCIL_BUFFER_BIT);
						}

						// Clip against other primitives
						// in the product.
					
						int bit = 0;
						unsigned int mask = 0;

						for (int q=0; q<sum[t].size(); q++)
						{
							if (q!=p)
							{
								if (sum[t].elementComplemented(q))
									sum[t].element(q)->subtract(bit,mask);
								else
									sum[t].element(q)->intersect(bit,mask);
								bit++;
							}

							if ( (bit==stencilSize || q==sum[t].size()-1) && bit )
							{
								parityApply(mask); // Mask out bits that fail parity test
								glClear(GL_STENCIL_BUFFER_BIT);
								mask = 0;
								bit = 0;
							}
						}

						// Merge with saved z-buffer, if necessary

						if (firstTime)
							firstTime = false;
						else
							writeDepthBuffer();
					}
			}
		}

		//
		// Use final z-buffer to match primitive surfaces
		//

		glDisable(GL_STENCIL_TEST);
		glEnable(GL_DEPTH_TEST);
		glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
		glEnable(GL_LIGHTING);
		glEnable(GL_NORMALIZE);
		glDepthFunc(GL_EQUAL);
		glDepthMask(GL_FALSE);
		
		for (Uint prod=0; prod<sum.size(); prod++)
			for (Uint prim=0; prim<sum[prod].size(); prim++)
				if (sum[prod].elementComplemented(prim))
					sum[prod].element(prim)->drawBack();
				else
					sum[prod].element(prim)->drawFront();
			
	glPopAttrib();
}

//
// Z-Buffer management
//

void readDepthBuffer()
{
	if (useCopy)
	{
		glPushAttrib(GL_VIEWPORT_BIT | GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

			glViewport(0,0,viewport[2],viewport[3]);

			glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glLoadIdentity();
			glOrtho(0,viewport[2],0,viewport[3],0.0,1.0);

			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();
			glLoadIdentity();
			glRasterPos2f(viewport[2]>>1,0.0F);

			glDisable(GL_STENCIL_TEST);
			glEnable(GL_DEPTH_TEST);
			glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
			glDepthMask(GL_TRUE);
			glDepthFunc(GL_ALWAYS);
			glDisable(GL_CULL_FACE);
			
			glCopyPixels(0,0,viewport[2]>>1,viewport[3],GL_DEPTH);

			glPopMatrix();
			glMatrixMode(GL_PROJECTION);
			glPopMatrix();
			glMatrixMode(GL_MODELVIEW);

		glPopAttrib();
	}
	else
		depthCopy.read();
}

void writeDepthBuffer()
{
	glDisable(GL_STENCIL_TEST);
	glEnable(GL_DEPTH_TEST);
	glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
	glDepthMask(GL_TRUE);
	glDepthFunc(GL_LESS);
	glDisable(GL_CULL_FACE);

	if (useCopy)
	{
		glPushAttrib(GL_VIEWPORT_BIT);

			glViewport(0,0,viewport[2],viewport[3]);

			glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glLoadIdentity();
			glOrtho(0,viewport[2],0,viewport[3],0.0,1.0);

			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();
			glLoadIdentity();
			glRasterPos2f(0.0F,0.0F);

			glCopyPixels(viewport[2]>>1,0,viewport[2]>>1,viewport[3],GL_DEPTH);

			glPopMatrix();
			glMatrixMode(GL_PROJECTION);
			glPopMatrix();
			glMatrixMode(GL_MODELVIEW);

		glPopAttrib();
	}
	else
		depthCopy.write();
}

void parityApply(const int mask)
{
	glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

		glEnable(GL_STENCIL_TEST);
		glStencilFunc(GL_NOTEQUAL,mask,0xffff);    
		glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
		glStencilMask(0);

		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_ALWAYS);
		glDepthMask(GL_TRUE);
		glDisable(GL_CULL_FACE);

		drawZfar();

	glPopAttrib();
}

