//
// OpenGL CSG Rendering
//
// Nigel Stewart (nigels@eisa.net.au)
// RMIT Department of Mechanical and Manufacturing Engineering
// 27th December 1998
//
// LayeredGoldfeatherRenderingAlgorithm
//
// This algorithm is documented in detail in the
// following paper:
//
// "An Improved CSG Rendering Algorithm"
// N. Stewart, G. Leach, S. John
// 1998 Eurographics/Siggraph Graphics Hardware Workshop
// Lisbon, Portugal.
//
//
// The idea here is to clip more surfaces at once by 
// advantage of depth-complexity in the scene.  
// For front and back facing surfaces, we clip each 'layer' 
// rather than each primitive.  The assumption is that 
// the degree of overlap is less than n^2.  
// If depth-complexity is denoted 'k', this is O(kn)
// rather than O(n^2).
//

#include <goldlayr.h>
#include <goldfeat.h>

#include <zplane.h>

GLubyte 
countSurfaces(const vector<CsgTreeLeaf *> &leaf,const countSurfacesMode mode)
{
	GLuint dl = glGenLists(1);
	glNewList(dl,GL_COMPILE);
	for (int p=0; p<leaf.size(); p++)
		leaf[p]->draw();
	glEndList();

	GLubyte count = ::countSurfaces(&dl,1,mode);
	glDeleteLists(dl,1);

	return count;
}

void LayeredGoldfeatherRenderingAlgorithm(const CsgSumOfProducts &sum,bool copy)
{
	// A fakey trick that shouldn't be necessary
	//
	// WindowsNT software OpenGL seems to lose
	// precision on transformation matricies when
	// they are pushed/popped.
	//

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glPopMatrix();

	//
	// Z-Buffer management specified by caller.
	// Copy uses glCopyPixels
	// otherwise, use glReadPixels/glDrawPixels

	useCopy = copy;

	// Determine width of stencil buffer

	GLint		stencilSize = 0;
	glGetIntegerv(GL_STENCIL_BITS,&stencilSize);

	// Save our state so we leave things tidy for next time

	glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT | GL_POLYGON_BIT | GL_VIEWPORT_BIT);

		// Copy-pixels variant depends on viewport subdivision

		glGetIntegerv(GL_VIEWPORT,(GLint *) viewport);
		if (useCopy)
			glViewport(0,0,viewport[2]>>1,viewport[3]);

		// We handle each product seperately

		for (int t=0; t<sum.size(); t++)
		{
			// Preserve current z-buffer after first layer clipped

			bool firstTime = t==0;

			if (firstTime)
				readDepthBuffer();

			// Solution is formed in z-buffer, we draw colour later

			glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
			glDisable(GL_LIGHTING);
			glDisable(GL_NORMALIZE);

			// There is no need to clip anything if the
			// product consists of only one primitive.

			if (sum[t].size()==1)
			{
				// If this is not first product there may be clipped
				// surfaces we want to merge later

				glDisable(GL_STENCIL_TEST);

				glEnable(GL_DEPTH_TEST);
				glDepthFunc(GL_LESS);
				glDepthMask(GL_TRUE);

				// Loop through products while size is 1
				// and draw front surfaces unconditionally

				while (t<sum.size() && sum[t].size()==1)
					sum[t++].element(0)->drawFront();

				// Here we merge previously clipped surfaces

				if (firstTime)
					firstTime = false;
				else
					writeDepthBuffer();
			}
			else
			{
				// Here's the variation on Goldfeather.
				// Start with front facing surfaces of non-inverted
				// primitives.  Extract the 1 to kth layer and
				// clip against _all_ primitives.

				glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);

				int kFront = countSurfaces(sum[t].uncomplemented(),COUNT_SURFACES_99TH_PERCENTILE);
				int kBack = countSurfaces(sum[t].complemented(),COUNT_SURFACES_99TH_PERCENTILE);

					#ifdef _DEBUG
					cout << kFront << " front facing surfaces." << endl;
					cout << kBack  << " back facing surfaces." << endl;
					#endif

				assert(kFront<(1<<stencilSize));

				for (int k=0; k<kFront+kBack; k++)
				{
					if (k>0)
						readDepthBuffer();
#if 0
					glClear(GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
#else
					glClear(GL_STENCIL_BUFFER_BIT);
#endif

					// Depth test for parity depends on front or back
					//
					// For front facing, count front face as being inside.
					// For back facing, count back face as being outside.
					
					GLenum depthTest = (k<kFront ? GL_LEQUAL : GL_LEQUAL );

					// Draw k'th layer into z-buffer

					glDepthFunc(GL_ALWAYS);
					glDepthMask(GL_TRUE);
					glEnable(GL_CULL_FACE);

					glEnable(GL_STENCIL_TEST);
					glStencilOp(GL_INCR,GL_INCR,GL_INCR);
					glStencilMask(0xffff);

					if (k<kFront)
					{
						glCullFace(GL_BACK);
						glStencilFunc(GL_EQUAL,k,0xffff);
						for (int p=0; p<sum[t].uncomplemented().size(); p++)
							sum[t].uncomplemented()[p]->draw();
					}
					else
					{
						glCullFace(GL_FRONT);
						glStencilFunc(GL_EQUAL,k-kFront,0xffff);
						for (int p=0; p<sum[t].complemented().size(); p++)
							sum[t].complemented()[p]->draw();
					}

					// Parity test layer against all primitives

					unsigned int mask = 0;
					int bit = 0;

					glClear(GL_STENCIL_BUFFER_BIT);

					for (int p=0; p<sum[t].size(); p++)
					{
						if (sum[t].elementComplemented(p))
							sum[t].element(p)->subtract(bit,mask,depthTest);
						else
							sum[t].element(p)->intersect(bit,mask,depthTest);

						bit++;

						if ( (bit==stencilSize || p==sum[t].size()-1) && bit )
						{
							parityApply(mask); // Mask out bits that fail parity test
							glClear(GL_STENCIL_BUFFER_BIT);
							mask = 0;
							bit = 0;
						}
					}

					writeDepthBuffer();
				}
			}
		}

		//
		// Use final z-buffer to match primitive surfaces
		//

		glDisable(GL_STENCIL_TEST);
		glEnable(GL_DEPTH_TEST);
		glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
		glEnable(GL_LIGHTING);
		glEnable(GL_NORMALIZE);
		glDepthFunc(GL_EQUAL);
		glDepthMask(GL_FALSE);
		
		for (Uint prod=0; prod<sum.size(); prod++)
			for (Uint prim=0; prim<sum[prod].size(); prim++)
				if (sum[prod].elementComplemented(prim))
					sum[prod].element(prim)->drawBack();
				else
					sum[prod].element(prim)->drawFront();
						
	glPopAttrib();
}

