#include <matrix4.h>

#include <math.h>
#include <iomanip.h>
#include <string.h>


double Matrix::_identity[16] = 
{
   1.0, 0.0, 0.0, 0.0,
   0.0, 1.0, 0.0, 0.0,
   0.0, 0.0, 1.0, 0.0,
   0.0, 0.0, 0.0, 1.0
};


Matrix::Matrix()
{
	memcpy(_matrix,_identity,16*sizeof(double));
}

Matrix::Matrix(const Matrix &matrix)
{
	(*this) = matrix;
}

Matrix::Matrix(double *matrix)
{
	memcpy(_matrix,matrix,sizeof(_matrix));
}

Matrix::Matrix(GLenum matrix)
{
	switch (matrix)
	{
	case GL_MODELVIEW_MATRIX:
		glGetDoublev(GL_MODELVIEW_MATRIX,_matrix);
		break;
	case GL_PROJECTION_MATRIX:
		glGetDoublev(GL_PROJECTION_MATRIX,_matrix);
		break;
	default:
		break;
	}
}

Matrix &
Matrix::operator=(const Matrix &m)
{
	memcpy(_matrix,m._matrix,sizeof(_matrix));

	return *this;
}

Matrix 
Matrix::operator *(const Matrix &m) const
{
	Matrix prod;

	for (int c=0;c<4;c++)
		for (int r=0;r<4;r++)
			prod.set(c,r,
				get(c,0)*m.get(0,r) +
				get(c,1)*m.get(1,r) +
				get(c,2)*m.get(2,r) +
				get(c,3)*m.get(3,r));

	return prod;
}

Vector    
Matrix::operator*(const Vector &v) const
{
	double prod[4] = { 0,0,0,0 };

	for (int r=0;r<4;r++)
	{
		for (int c=0;c<3;c++)
			prod[r] += v[c]*get(c,r);

		prod[r] += get(3,r);
	}

	double div = 1.0 / prod[3];

	return Vector(prod[0]*div,prod[1]*div,prod[2]*div);
}

double 
Matrix::operator[](const int index) const
{
	return _matrix[index];
}

bool 
Matrix::operator==(const Matrix &m) const
{
	return !memcmp(_matrix,m._matrix,16*sizeof(double));
}

bool 
Matrix::operator!=(const Matrix &m) const
{
	return memcmp(_matrix,m._matrix,16*sizeof(double))!=0;
}

Matrix::operator double *()
{
	return _matrix;
}

void
Matrix::glMultMatrix()
{
	glMultMatrixd(_matrix);
}

void 
Matrix::glLoadMatrix()
{
	glLoadMatrixd(_matrix);
}


#if 0
double const *
Matrix::row(const unsigned int row) const
{
	static double r[4];

	for (int c=0;c<4;c++)
		r[c] = _matrix[c*4+row];
	
	return r;
}

double const *
Matrix::column(const unsigned int column) const
{
	return _matrix+column*4;
}

#endif

ostream &
operator<<(ostream &os,const Matrix &m)
{
	for (int r=0;r<4;r++)
		for (int c=0;c<4;c++)
			os << setw(10) << setfill(' ') << m.get(c,r) << ( c==3 ? "\n" : ", ");

	return os;
}

ostream &
Matrix::writePOV(ostream &os) const
{
	UnMatrix um = unmatrix();

	os << "scale < ";
	  os << um[U_SCALEX] << ',';
	  os << um[U_SCALEY] << ',';
	  os << um[U_SCALEZ] << " > ";

  	os << "rotate < ";
	  os << um[U_ROTATEX]/Degr << ',';
	  os << um[U_ROTATEY]/Degr << ',';
	  os << um[U_ROTATEZ]/Degr << " > ";

  	os << "translate < ";
	  os << um[U_TRANSX] << ',';
	  os << um[U_TRANSY] << ',';
	  os << um[U_TRANSZ] << " > ";

	return os;
}

Matrix matrixScale(const double sf)
{
	Matrix scale;

	scale.set(0,0,sf);
	scale.set(1,1,sf);
	scale.set(2,2,sf);
	
	return scale;
}

Matrix matrixScale(const Vector sf)
{
	Matrix scale;

	scale.set(0,0,sf[0]);
	scale.set(1,1,sf[1]);
	scale.set(2,2,sf[2]);
	
	return scale;
}


Matrix matrixTranslate(const Vector trans)
{
	Matrix translate;

	translate.set(3,0,trans[0]);
	translate.set(3,1,trans[1]);
	translate.set(3,2,trans[2]);

	return translate;
}

Matrix matrixRotate(const Vector axis, const double angle)
{
	Matrix rotate;

	// Page 466, Graphics Gems

	double s = sin(angle*Pi/180.0);
	double c = cos(angle*Pi/180.0);
	double t = 1 - c;

	Vector ax = axis/sqrt(axis.norm());

	double x = ax[0];
	double y = ax[1];
	double z = ax[2];

	rotate.set(0,0,t*x*x+c);
	rotate.set(1,0,t*y*x+s*z);
	rotate.set(2,0,t*z*x-s*y);

	rotate.set(0,1,t*x*y-s*z);
	rotate.set(1,1,t*y*y+c);
	rotate.set(2,1,t*z*y+s*x);

	rotate.set(0,2,t*x*z+s*y);
	rotate.set(1,2,t*y*z-s*x);
	rotate.set(2,2,t*z*z+c);

	return rotate;
}

Matrix matrixRotate(const double azimuth, const double elevation)
{
	Matrix rotate;

	double ca = cos(azimuth*Pi/180.0);
	double sa = sin(azimuth*Pi/180.0);
	double cb = cos(elevation*Pi/180.0);
	double sb = sin(elevation*Pi/180.0);

	rotate.set(0,0,cb);
	rotate.set(1,0,0);
	rotate.set(2,0,-sb);

	rotate.set(0,1,-sa*sb);
	rotate.set(1,1,ca);
	rotate.set(2,1,-sa*cb);

	rotate.set(0,2,ca*sb);
	rotate.set(1,2,sa);
	rotate.set(2,2,ca*cb);

	return rotate;
}


Matrix 
Matrix::inverse() 
{
	Matrix inv;
	invertMatrix(_matrix,inv._matrix);
	return inv;
}

Matrix 
Matrix::transpose()
{
	Matrix tmp;
	
	for (int i=0;i<4;i++)
		for (int j=0;j<4;j++)
			tmp.set(j,i,get(i,j));
	
	return tmp;
}

//
// From Mesa-2.2\src\glu\project.c
//

//
// Compute the inverse of a 4x4 matrix.  Contributed by scotter@lafn.org
//

void 
Matrix::invertMatrixGeneral( const GLdouble *m, GLdouble *out )
{

/* NB. OpenGL Matrices are COLUMN major. */
#define MAT(m,r,c) (m)[(c)*4+(r)]

/* Here's some shorthand converting standard (row,column) to index. */
#define m11 MAT(m,0,0)
#define m12 MAT(m,0,1)
#define m13 MAT(m,0,2)
#define m14 MAT(m,0,3)
#define m21 MAT(m,1,0)
#define m22 MAT(m,1,1)
#define m23 MAT(m,1,2)
#define m24 MAT(m,1,3)
#define m31 MAT(m,2,0)
#define m32 MAT(m,2,1)
#define m33 MAT(m,2,2)
#define m34 MAT(m,2,3)
#define m41 MAT(m,3,0)
#define m42 MAT(m,3,1)
#define m43 MAT(m,3,2)
#define m44 MAT(m,3,3)

   GLdouble det;
   GLdouble d12, d13, d23, d24, d34, d41;
   GLdouble tmp[16]; /* Allow out == in. */

   /* Inverse = adjoint / det. (See linear algebra texts.)*/

   /* pre-compute 2x2 dets for last two rows when computing */
   /* cofactors of first two rows. */
   d12 = (m31*m42-m41*m32);
   d13 = (m31*m43-m41*m33);
   d23 = (m32*m43-m42*m33);
   d24 = (m32*m44-m42*m34);
   d34 = (m33*m44-m43*m34);
   d41 = (m34*m41-m44*m31);

   tmp[0] =  (m22 * d34 - m23 * d24 + m24 * d23);
   tmp[1] = -(m21 * d34 + m23 * d41 + m24 * d13);
   tmp[2] =  (m21 * d24 + m22 * d41 + m24 * d12);
   tmp[3] = -(m21 * d23 - m22 * d13 + m23 * d12);

   /* Compute determinant as early as possible using these cofactors. */
   det = m11 * tmp[0] + m12 * tmp[1] + m13 * tmp[2] + m14 * tmp[3];

   /* Run singularity test. */
   if (det == 0.0) {
      /* printf("invert_matrix: Warning: Singular matrix.\n"); */
	  memcpy(_matrix,_identity,16*sizeof(double));
   }
   else {
      GLdouble invDet = 1.0 / det;
      /* Compute rest of inverse. */
      tmp[0] *= invDet;
      tmp[1] *= invDet;
      tmp[2] *= invDet;
      tmp[3] *= invDet;

      tmp[4] = -(m12 * d34 - m13 * d24 + m14 * d23) * invDet;
      tmp[5] =  (m11 * d34 + m13 * d41 + m14 * d13) * invDet;
      tmp[6] = -(m11 * d24 + m12 * d41 + m14 * d12) * invDet;
      tmp[7] =  (m11 * d23 - m12 * d13 + m13 * d12) * invDet;

      /* Pre-compute 2x2 dets for first two rows when computing */
      /* cofactors of last two rows. */
      d12 = m11*m22-m21*m12;
      d13 = m11*m23-m21*m13;
      d23 = m12*m23-m22*m13;
      d24 = m12*m24-m22*m14;
      d34 = m13*m24-m23*m14;
      d41 = m14*m21-m24*m11;

      tmp[8] =  (m42 * d34 - m43 * d24 + m44 * d23) * invDet;
      tmp[9] = -(m41 * d34 + m43 * d41 + m44 * d13) * invDet;
      tmp[10] =  (m41 * d24 + m42 * d41 + m44 * d12) * invDet;
      tmp[11] = -(m41 * d23 - m42 * d13 + m43 * d12) * invDet;
      tmp[12] = -(m32 * d34 - m33 * d24 + m34 * d23) * invDet;
      tmp[13] =  (m31 * d34 + m33 * d41 + m34 * d13) * invDet;
      tmp[14] = -(m31 * d24 + m32 * d41 + m34 * d12) * invDet;
      tmp[15] =  (m31 * d23 - m32 * d13 + m33 * d12) * invDet;

      memcpy(out, tmp, 16*sizeof(GLdouble));
   }

#undef m11
#undef m12
#undef m13
#undef m14
#undef m21
#undef m22
#undef m23
#undef m24
#undef m31
#undef m32
#undef m33
#undef m34
#undef m41
#undef m42
#undef m43
#undef m44
#undef MAT
}


//
// Invert matrix m.  This algorithm contributed by Stephane Rehel
// <rehel@worldnet.fr>
//

void 
Matrix::invertMatrix( const GLdouble *m, GLdouble *out )
{
/* NB. OpenGL Matrices are COLUMN major. */
#define MAT(m,r,c) (m)[(c)*4+(r)]

/* Here's some shorthand converting standard (row,column) to index. */
#define m11 MAT(m,0,0)
#define m12 MAT(m,0,1)
#define m13 MAT(m,0,2)
#define m14 MAT(m,0,3)
#define m21 MAT(m,1,0)
#define m22 MAT(m,1,1)
#define m23 MAT(m,1,2)
#define m24 MAT(m,1,3)
#define m31 MAT(m,2,0)
#define m32 MAT(m,2,1)
#define m33 MAT(m,2,2)
#define m34 MAT(m,2,3)
#define m41 MAT(m,3,0)
#define m42 MAT(m,3,1)
#define m43 MAT(m,3,2)
#define m44 MAT(m,3,3)

   register GLdouble det;
   GLdouble tmp[16]; /* Allow out == in. */

   if( m41 != 0. || m42 != 0. || m43 != 0. || m44 != 1. ) {
      invertMatrixGeneral(m, out);
      return;
   }

   /* Inverse = adjoint / det. (See linear algebra texts.)*/

   tmp[0]= m22 * m33 - m23 * m32;
   tmp[1]= m23 * m31 - m21 * m33;
   tmp[2]= m21 * m32 - m22 * m31;

   /* Compute determinant as early as possible using these cofactors. */
   det= m11 * tmp[0] + m12 * tmp[1] + m13 * tmp[2];

   /* Run singularity test. */
   if (det == 0.0) {
      /* printf("invert_matrix: Warning: Singular matrix.\n"); */
      memcpy( out, _identity, 16*sizeof(GLdouble) );
   }
   else {
      GLdouble d12, d13, d23, d24, d34, d41;
      register GLdouble im11, im12, im13, im14;

      det= 1. / det;

      /* Compute rest of inverse. */
      tmp[0] *= det;
      tmp[1] *= det;
      tmp[2] *= det;
      tmp[3]  = 0.;

      im11= m11 * det;
      im12= m12 * det;
      im13= m13 * det;
      im14= m14 * det;
      tmp[4] = im13 * m32 - im12 * m33;
      tmp[5] = im11 * m33 - im13 * m31;
      tmp[6] = im12 * m31 - im11 * m32;
      tmp[7] = 0.;

      /* Pre-compute 2x2 dets for first two rows when computing */
      /* cofactors of last two rows. */
      d12 = im11*m22 - m21*im12;
      d13 = im11*m23 - m21*im13;
      d23 = im12*m23 - m22*im13;
      d24 = im12*m24 - m22*im14;
      d34 = im13*m24 - m23*im14;
      d41 = im14*m21 - m24*im11;

      tmp[8] =  d23;
      tmp[9] = -d13;
      tmp[10] = d12;
      tmp[11] = 0.;

      tmp[12] = -(m32 * d34 - m33 * d24 + m34 * d23);
      tmp[13] =  (m31 * d34 + m33 * d41 + m34 * d13);
      tmp[14] = -(m31 * d24 + m32 * d41 + m34 * d12);
      tmp[15] =  1.;

      memcpy(out, tmp, 16*sizeof(GLdouble));
  }

#undef m11
#undef m12
#undef m13
#undef m14
#undef m21
#undef m22
#undef m23
#undef m24
#undef m31
#undef m32
#undef m33
#undef m34
#undef m41
#undef m42
#undef m43
#undef m44
#undef MAT
}

//
// From Graphics Gems II - Decomposing a matrix into simple transformations. Pg. 320
//

// unmatrix.c - given a 4x4 matrix, decompose it into standard operations.
//
// Author:	Spencer W. Thomas
//          University of Michigan
//

// unmatrix - Decompose a non-degenerate 4x4 transformation matrix into
// the sequence of transformations that produced it.
// [Sx][Sy][Sz][Shearx/y][Sx/z][Sz/y][Rx][Ry][Rz][Tx][Ty][Tz][P(x,y,z,w)]
//
// The coefficient of each transformation is returned in the corresponding
// element of the vector tran.
//
// Returns true upon success, false if the matrix is singular.
//

UnMatrix::UnMatrix()
{
	for (int f=0;f<16;f++)
		_tran[f] = 0.0;
}


static char *UnMatrixFieldDescription[] =
{
 	"U_SCALEX",
 	"U_SCALEY",
 	"U_SCALEZ",
 	"U_SHEARXY",
 	"U_SHEARXZ",
 	"U_SHEARYZ",
 	"U_ROTATEX",
 	"U_ROTATEY",
 	"U_ROTATEZ",
 	"U_TRANSX",
 	"U_TRANSY",
 	"U_TRANSZ",
 	"U_PERSPX",
 	"U_PERSPY",
 	"U_PERSPZ",
 	"U_PERSPW"
};

ostream &
operator<<(ostream &os,const UnMatrixField &field)
{
	if (field>=U_SCALEX && field<=U_PERSPW)
		os << UnMatrixFieldDescription[field];
	else
		os << "Unknown";

	return os;
}

ostream &
operator<<(ostream &os,const UnMatrix &unMatrix)
{
	for (int f=0;f<16;f++)
		os << (UnMatrixField) f << ' ' << unMatrix._tran[f] << endl;

	return os;
}

istream &
operator>>(istream &is,UnMatrixField &field)
{
	char buffer[128];

	is >> buffer;

	for (int f=0;f<16;f++)
		if (!strcmp(buffer,UnMatrixFieldDescription[f]))
		{
			field = (UnMatrixField) f;
			break;
		}

	return is;
}

istream &
operator>>(istream &is,UnMatrix &unMatrix)
{
	for (int f=0;f<16;f++)
	{
		UnMatrixField field;
		double value;

		is >> field;
		is >> value;

		if (!is.eof()) unMatrix[field] = value;
	}

	return is;
}

typedef struct {
	double x,y,z,w;
} Vector4;

UnMatrix
Matrix::unmatrix() const
{
	UnMatrix tran;

 	Matrix locmat = (*this);

	if (locmat.element(3,3)==0.0)
		return tran;

 	register int i, j;

	for ( i=0; i<4;i++ )
 		for ( j=0; j<4; j++ )
 			locmat.element(i,j) /= locmat.element(3,3);

 	/* pmat is used to solve for perspective, but it also provides
 	 * an easy way to test for singularity of the upper 3x3 component.
 	 */

 	Matrix pmat = locmat;

 	for ( i=0; i<3; i++ )
 		pmat.set(i,3,0.0);

 	pmat.set(3,3,1.0);

 	if ( pmat.det() == 0.0 )
 		return tran;

 	/* First, isolate perspective.  This is the messiest. */

 	if 
	( 
		locmat.element(0,3) != 0.0 || 
		locmat.element(1,3) != 0.0 ||
 		locmat.element(2,3) != 0.0 
	) 
	{
		Vector4 prhs, psol;

 		/* prhs is the right hand side of the equation. */
 
		prhs.x = locmat.element(0,3);
 		prhs.y = locmat.element(1,3);
 		prhs.z = locmat.element(2,3);
 		prhs.w = locmat.element(3,3);

 		/* Solve the equation by inverting pmat and multiplying
 		 * prhs by the inverse.  (This is the easiest way, not
 		 * necessarily the best.)
 		 * inverse function (and det4x4, above) from the Matrix
 		 * Inversion gem in the first volume.
 		 */

//		Matrix invpmat = pmat.inverse();
//		Matrix tinvpmat = invpmat.transpose();
// 		V4MulPointByMatrix(&prhs, &tinvpmat, &psol);    // REVISIT
 
		psol.x = 0.0;
		psol.y = 0.0;
		psol.z = 0.0;
		psol.w = 0.0;

 		/* Stuff the answer away. */
 		tran[U_PERSPX] = psol.x;
 		tran[U_PERSPY] = psol.y;
 		tran[U_PERSPZ] = psol.z;
 		tran[U_PERSPW] = psol.w;

 		/* Clear the perspective partition. */
 		locmat.element(0,3) = locmat.element(1,3) =	locmat.element(2,3) = 0.0;
 		locmat.element(3,3) = 1.0;

 	} else		/* No perspective. */
		tran[U_PERSPX] = tran[U_PERSPY] = tran[U_PERSPZ] = tran[U_PERSPW] = 0;

 	/* Next take care of translation (easy). */
 	for ( i=0; i<3; i++ ) 
	{
 		tran[(UnMatrixField) (U_TRANSX + i)] = locmat.element(3,i);
 		locmat.set(3,i,0.0);
 	}

 	Vector  row[3];

 	/* Now get scale and shear. */
 	for ( i=0; i<3; i++ ) 
	{
 		row[i][0] = locmat.get(i,0);
 		row[i][1] = locmat.get(i,1);
 		row[i][2] = locmat.get(i,2);
 	}

 	/* Compute X scale factor and normalize first row. */

 	tran[U_SCALEX] = sqrt(row[0].norm());
 	row[0].normalise();

 	/* Compute XY shear factor and make 2nd row orthogonal to 1st. */
 	tran[U_SHEARXY] = row[0] * row[1];
	row[1] = row[1] - row[0] * tran[U_SHEARXY];

 	/* Now, compute Y scale and normalize 2nd row. */
 	tran[U_SCALEY] = sqrt(row[1].norm());
 	row[1].normalise();
 	tran[U_SHEARXY] /= tran[U_SCALEY];

 	/* Compute XZ and YZ shears, orthogonalize 3rd row. */
 	tran[U_SHEARXZ] = row[0] * row[2];
 	row[2] = row[2] - row[0] * tran[U_SHEARXZ];

 	tran[U_SHEARYZ] = row[1] * row[2];
	row[2] = row[2] - row[1] * tran[U_SHEARYZ];

 	/* Next, get Z scale and normalize 3rd row. */
 	tran[U_SCALEZ] = sqrt(row[2].norm());
 	row[2].normalise();

 	tran[U_SHEARXZ] /= tran[U_SCALEZ];
 	tran[U_SHEARYZ] /= tran[U_SCALEZ];
 
 	/* At this point, the matrix (in rows[]) is orthonormal.
 	 * Check for a coordinate system flip.  If the determinant
 	 * is -1, then negate the matrix and the scaling factors.
 	 */
 	if ( row[0] * xProduct(row[1],row[2]) < 0.0 )
 		for ( i = 0; i < 3; i++ ) 
		{
 			tran[(UnMatrixField) (U_SCALEX+i)] *= -1;
 			row[i] = row[i] * -1.0;
 		}
 
 	/* Now, get the rotations out, as described in the gem. */
 	tran[U_ROTATEY] = asin(-row[0][2]);

 	if ( cos(tran[U_ROTATEY]) != 0 ) 
	{
 		tran[U_ROTATEX] = atan2(row[1][2], row[2][2]);
 		tran[U_ROTATEZ] = atan2(row[0][1], row[0][0]);
 	} 
	else 
	{
 		tran[U_ROTATEX] = atan2(row[1][0], row[1][1]);
 		tran[U_ROTATEZ] = 0;
 	}

	return tran;
}

//
// From Graphics Gems GEMSI\MATINV.C
//

//
// Calculate the determinant of a 4x4 matrix.
//

double
Matrix::det() const
{
    double ans;
    double a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4, d1, d2, d3, d4;

    /* assign to individual variable names to aid selecting */
	/*  correct elements */

	a1 = get(0,0); b1 = get(0,1); 
	c1 = get(0,2); d1 = get(0,3);

	a2 = get(1,0); b2 = get(1,1); 
	c2 = get(1,2); d2 = get(1,3);

	a3 = get(2,0); b3 = get(2,1); 
	c3 = get(2,2); d3 = get(2,3);

	a4 = get(3,0); b4 = get(3,1); 
	c4 = get(3,2); d4 = get(3,3);

    ans = a1 * det3x3( b2, b3, b4, c2, c3, c4, d2, d3, d4)
        - b1 * det3x3( a2, a3, a4, c2, c3, c4, d2, d3, d4)
        + c1 * det3x3( a2, a3, a4, b2, b3, b4, d2, d3, d4)
        - d1 * det3x3( a2, a3, a4, b2, b3, b4, c2, c3, c4);
    return ans;
}

// 
// Calculate the determinant of a 3x3 matrix
// in the form
//
//     | a1,  b1,  c1 |
//     | a2,  b2,  c2 |
//     | a3,  b3,  c3 |
//

double 
Matrix::det3x3
( 
	const double a1, 
	const double a2, 
	const double a3, 
	const double b1, 
	const double b2, 
	const double b3, 
	const double c1, 
	const double c2, 
	const double c3 
) const
{
    return 
		a1 * det2x2( b2, b3, c2, c3 )
        - b1 * det2x2( a2, a3, c2, c3 )
        + c1 * det2x2( a2, a3, b2, b3 );
}

// Calculate the determinant of a 2x2 matrix.

double
Matrix::det2x2
( 
	const double a, 
	const double b, 
	const double c, 
	const double d
) const
{
	return a * d - b * c;
}








UnMatrix 
operator-(const UnMatrix &b,const UnMatrix &a)
{
	UnMatrix um;

	um[U_SCALEX]  = b._tran[U_SCALEX] - a._tran[U_SCALEX];
	um[U_SCALEY]  = b._tran[U_SCALEY] - a._tran[U_SCALEY];
	um[U_SCALEZ]  = b._tran[U_SCALEZ] - a._tran[U_SCALEZ];

	um[U_ROTATEX] = b._tran[U_ROTATEX] - a._tran[U_ROTATEX];
	um[U_ROTATEY] = b._tran[U_ROTATEY] - a._tran[U_ROTATEY];
	um[U_ROTATEZ] = b._tran[U_ROTATEZ] - a._tran[U_ROTATEZ];

	um[U_TRANSX]  = b._tran[U_TRANSX] - a._tran[U_TRANSX];
	um[U_TRANSY]  = b._tran[U_TRANSY] - a._tran[U_TRANSY];
	um[U_TRANSZ]  = b._tran[U_TRANSZ] - a._tran[U_TRANSZ];

	return um;
}

UnMatrix 
operator*(const UnMatrix &a,const double scaleFactor)
{
	UnMatrix um = a;

	um[U_SCALEX]  *= scaleFactor;
	um[U_SCALEY]  *= scaleFactor;
	um[U_SCALEZ]  *= scaleFactor;

	um[U_ROTATEX] *= scaleFactor;
	um[U_ROTATEY] *= scaleFactor;
	um[U_ROTATEZ] *= scaleFactor;

	um[U_TRANSX]  *= scaleFactor;
	um[U_TRANSY]  *= scaleFactor;
	um[U_TRANSZ]  *= scaleFactor;

	return um;
}

UnMatrix 
operator+(const UnMatrix &a,const UnMatrix &b)
{
	UnMatrix um;

	um[U_SCALEX]  = b._tran[U_SCALEX] + a._tran[U_SCALEX];
	um[U_SCALEY]  = b._tran[U_SCALEY] + a._tran[U_SCALEY];
	um[U_SCALEZ]  = b._tran[U_SCALEZ] + a._tran[U_SCALEZ];

	um[U_ROTATEX] = b._tran[U_ROTATEX] + a._tran[U_ROTATEX];
	um[U_ROTATEY] = b._tran[U_ROTATEY] + a._tran[U_ROTATEY];
	um[U_ROTATEZ] = b._tran[U_ROTATEZ] + a._tran[U_ROTATEZ];

	um[U_TRANSX]  = b._tran[U_TRANSX] + a._tran[U_TRANSX];
	um[U_TRANSY]  = b._tran[U_TRANSY] + a._tran[U_TRANSY];
	um[U_TRANSZ]  = b._tran[U_TRANSZ] + a._tran[U_TRANSZ];

	return um;
}



