//
// OpenGL Solid Modelling
//
// Nigel Stewart (nigels@eisa.net.au)
// RMIT Department of Mechanical and Manufacturing Engineering
// 11th December 1997
//

#ifndef MATRIX4_H
#define MATRIX4_H

#include <atypes.h>
#include <vector3.h>
#include <iostream.h>

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>

class UnMatrix;

////////////////////////// Matrix /////////////////////////////////

class Matrix
{
	friend Matrix matrixScale(const double sf);
	friend Matrix matrixScale(const Vector sf);
	friend Matrix matrixTranslate(const Vector trans);
	friend Matrix matrixRotate(const Vector axis,const double angle);
	friend Matrix matrixRotate(const double azimuth,const double elevation);

	friend ostream &operator<<(ostream &os,const Matrix &m);

public:
	Matrix();
	Matrix(const Matrix &matrix);
	Matrix(double *matrix);
	Matrix(GLenum matrix);

	Matrix &operator=(const Matrix &);

	Matrix operator*(const Matrix &) const;
	Vector operator*(const Vector &) const;

	double operator[](const int index) const;

	bool operator==(const Matrix &) const;
	bool operator!=(const Matrix &) const;

	operator double * ();

	Matrix inverse();
	Matrix transpose();

	void glMultMatrix();
	void glLoadMatrix();

	ostream &writePOV(ostream &os) const;

	UnMatrix unmatrix() const;

	double det() const;

private:

	double _matrix[16];
	static double _identity[16];

	inline void set(const int col,const int row,const double val) 
	{ 
		_matrix[col*4+row] = val;
	}

	inline double get(const int col,const int row) const
	{
		return _matrix[col*4+row];
	}

	inline double &element(const int col,const int row) 
	{
		return _matrix[col*4+row];
	}

	// From Mesa-2.2\src\glu\project.c

	void invertMatrixGeneral( const GLdouble *m, GLdouble *out );
	void invertMatrix( const GLdouble *m, GLdouble *out );

	// From Graphics Gems GEMSI\MATINV.C

	double 
	det3x3
	( 
		const double a1, 
		const double a2, 
		const double a3, 
		const double b1, 
		const double b2, 
		const double b3, 
		const double c1, 
		const double c2, 
		const double c3 
	) const;

	double
	det2x2
	( 
		const double a, 
		const double b, 
		const double c, 
		const double d
	) const;
};

////////////////////////// UnMatrix /////////////////////////////////

//
// From Graphics Gems II - Decomposing a matrix into simple transformations. Pg. 320
//

// Author:	Spencer W. Thomas
//          University of Michigan
//

// The unmatrix subroutine fills in a vector of floating point
// values.  These symbols make it easier to get the data back out.
//

typedef enum
{
 	U_SCALEX = 0,
 	U_SCALEY,
 	U_SCALEZ,
 	U_SHEARXY,
 	U_SHEARXZ,
 	U_SHEARYZ,
 	U_ROTATEX,
 	U_ROTATEY,
 	U_ROTATEZ,
 	U_TRANSX,
 	U_TRANSY,
 	U_TRANSZ,
 	U_PERSPX,
 	U_PERSPY,
 	U_PERSPZ,
 	U_PERSPW
} UnMatrixField;

const int UnMatrixFieldNumber = 16;

class UnMatrix
{
	friend Matrix;

	friend ostream &operator<<(ostream &os,const UnMatrixField &field);
	friend ostream &operator<<(ostream &os,const UnMatrix &unMatrix);

	friend istream &operator>>(istream &is,UnMatrixField &field);
	friend istream &operator>>(istream &is,UnMatrix &unMatrix);

	friend UnMatrix operator-(const UnMatrix &b,const UnMatrix &a);
	friend UnMatrix operator*(const UnMatrix &a,const double scaleFactor);
	friend UnMatrix operator+(const UnMatrix &a,const UnMatrix &b);

public:
	UnMatrix();

	double &operator[](const UnMatrixField field) { return _tran[field]; };

private:
	double _tran[16];
};

#endif
