//
// OpenGL Solid Modelling
//
// Nigel Stewart (nigels@eisa.net.au)
// RMIT Department of Manufacturing Systems Engineering
// 26th December 1998
//
// SCS Algorithm
//

#include <scs.h>

#include <zplane.h>
#include <timer.h>
#include <goldfeat.h>

#include <assert.h>

vector<Uint> subtractionSequence(const Uint n, Uint k)
{
	// Check sanity of parameters.

	assert(k<=n);

	// If k is not supplied, assume
	// k is same as n.  This is the
	// worst case.

	if (k==0)
		k = n;

	// Store generated sequence.

	vector<Uint> sequence;

	// Main Loop

	for (Uint i=0; i<k; i++)
		for (Uint j=0; j<n; j++)
			if (i==0 || j>0)
				sequence.push_back( (i%2)==0 ? j : n-1-j );

	return sequence;
}

void outputSequence(ostream &os,const vector<Uint> &seq)
{
	for (Uint i=0; i<seq.size(); i++)
		os << char(CLAMP(Uint('a'+seq[i]),Uint('a'),Uint('z')));
}

GLubyte 
countSurfaces(const CsgProduct &prod,const countSurfacesMode mode)
{
	GLuint dl = glGenLists(1);
	glNewList(dl,GL_COMPILE);
	for (int p=1; p<prod.size(); p++)
			prod.element(p)->draw();
	glEndList();

	GLubyte count = ::countSurfaces(&dl,1,mode);
	glDeleteLists(dl,1);

	glClear(GL_STENCIL_BUFFER_BIT);

	return count;
}

//
// Rendering
//

void SequencedConvexSubtraction(const CsgSumOfProducts &sumOfProd,bool copy)
{
	glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT | GL_POLYGON_BIT | GL_VIEWPORT_BIT);

	//

	useCopy = copy;

	glGetIntegerv(GL_VIEWPORT,(GLint *) viewport);
	if (useCopy)
		glViewport(0,0,viewport[2]>>1,viewport[3]);

	Timer timer;

	// A fakey trick that shouldn't be necessary
	//
	// WindowsNT software OpenGL seems to lose
	// precision on transformation matricies when
	// they are pushed/popped.
	//

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glPopMatrix();

	//

	glDisable(GL_LIGHTING);
	glDisable(GL_NORMALIZE);
	glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
	glDepthMask(GL_TRUE);
	glStencilMask(0xffff);

	glDisable(GL_STENCIL_TEST);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_ALWAYS);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);

	// For each product

	for (Uint p=0; p<sumOfProd.size(); p++)
	{
		//

		const CsgProduct &prod = sumOfProd[p];

		//

		assert(prod.size()>0);

		Uint n = prod.size() - 1;

		if (n==0)
		{
			glDepthFunc(GL_LESS);
			prod.element(0)->draw();
			continue;
		}

		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);

		Uint k = countSurfaces(prod,COUNT_SURFACES_99TH_PERCENTILE);

		vector<Uint> seq = subtractionSequence(n,k);

		#ifdef _DEBUG
		cout << "p=" << p << ", n=" << n << ", k=" << k << " -> ";
		outputSequence(cout,seq);
		#endif

		// Draw front of initial volume into z-buffer

		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_ALWAYS);
		glDepthMask(GL_TRUE);

		prod.element(0)->draw();

		// Subtract according to sequence

		glEnable(GL_STENCIL_TEST);

		for (Uint i=0; i<seq.size(); i++)
		{
			// Prepare stencil buffer

			if (i>0)
				glClear(GL_STENCIL_BUFFER_BIT);

			Uint p = seq[i]+1;

			// Subtract a complemented object

			if (prod.elementComplemented(p))
			{
				glCullFace(GL_BACK);
				glDepthFunc(GL_LEQUAL);
				glDepthMask(GL_FALSE);
				glStencilFunc(GL_ALWAYS,1,0xffff);
				glStencilOp(GL_NONE,GL_NONE,GL_REPLACE);
				
				prod.element(p)->draw();

				glCullFace(GL_FRONT);
				glDepthFunc(GL_GREATER);
				glDepthMask(GL_TRUE);
				glStencilFunc(GL_EQUAL,1,0xffff);
				glStencilOp(GL_NONE,GL_NONE,GL_NONE);

				prod.element(p)->draw();
			}
			else
			// Subtract an un-complemented object
			{

				glCullFace(GL_BACK);
				glDepthFunc(GL_GREATER);
				glDisable(GL_STENCIL_TEST);

				prod.element(p)->draw();

				glCullFace(GL_FRONT);
				glDepthFunc(GL_LEQUAL);
				glDepthMask(GL_FALSE);
				glEnable(GL_STENCIL_TEST);
				glStencilFunc(GL_ALWAYS,1,1);
				glStencilOp(GL_KEEP,GL_REPLACE,GL_ZERO);
				
				prod.element(p)->draw();

				glDisable(GL_CULL_FACE);
				glDepthFunc(GL_ALWAYS);
				glDepthMask(GL_TRUE);
				glStencilFunc(GL_EQUAL,0,1);
				glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);

    			drawZfar();

				glEnable(GL_CULL_FACE);
			}
		}

		// Clip against back of initial volume

		glClear(GL_STENCIL_BUFFER_BIT);

		glCullFace(GL_FRONT);
		glDepthFunc(GL_LESS);
		glDepthMask(GL_FALSE);
		glStencilFunc(GL_ALWAYS,1,0xffff);
		glStencilOp(GL_KEEP,GL_KEEP,GL_REPLACE);

		prod.element(0)->draw();

		glDisable(GL_CULL_FACE);
		glDepthFunc(GL_ALWAYS);
		glDepthMask(GL_TRUE);
		glStencilFunc(GL_EQUAL,1,0xffff);
		glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);

		drawZfar();

		//
		// Z-Buffer management
		//
		// 1. Merge will current saved z-buffer, if necessary
		// 2. Copy z-buffer to safe place, if necessary
		//

		glDisable(GL_STENCIL_TEST);
		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_LESS);
		glDepthMask(GL_TRUE);

		if (p>0)
			writeDepthBuffer();

		if (p+1<sumOfProd.size())
		{
			readDepthBuffer();

			#ifdef _DEBUG
			cout << endl;
			#endif
		}
	}

	// Draw colour pixels

	glDepthFunc(GL_EQUAL);
	glDepthMask(GL_FALSE);
	glDisable(GL_STENCIL_TEST);
	glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
	glEnable(GL_LIGHTING);
	glEnable(GL_NORMALIZE);

	for (Uint product=0; product<sumOfProd.size(); product++)
	{
		const CsgProduct &prod = sumOfProd[product];
		for (Uint i=0; i<prod.size(); i++)
		{
			if (prod.elementComplemented(i))
				glCullFace(GL_FRONT);
			else
				glCullFace(GL_BACK);

			prod.element(i)->draw();
		}
	}

		#ifdef _DEBUG
		cout << ' ' << timer.elapsed() << " sec" << endl;
		#endif

	// ONYX2 Hack
#if 0
	glColor3f(0,0,0);
	glDisable(GL_LIGHTING);
	drawZfar();
#endif

	glPopAttrib();
}

