//
//
//

#include <sweep.h>

#include <math.h>
#include <assert.h>

//
// DisplayList version
//

GLuint Sweep(Func &f,const double radius,const int divisions,const int samples,const GLubyte *color)
{
	GLuint dl;
	dl = glGenLists(1);
	glNewList(dl,GL_COMPILE);

	if (color)
		glColor3ubv(color);

	Mesh *mesh = MeshSweep(f,radius,divisions,samples);
	
	if (mesh)
		mesh->glDraw();

	glEndList();

	delete mesh;

	return dl;
}

	
//
// Mesh version
//

vector<Mesh *> 
_MeshSweep(Func &f,const double radius,const int divisions,const int samples,const bool convex = false)
{
	Vector position, prev_position;
	Vector orientation, prev_orientation;
	Vector accel;
	Vector r;
	Vector *vertex = new Vector[samples];
	Vector *prev_vertex = new Vector[samples];

	vector<Mesh *> meshes;
	Mesh *mesh = new Mesh();

	int pos,vert;

	for (pos=-1;pos<=divisions;pos++)
	{
		// Get position and orientation

		position	= f.f((double) pos / divisions);
		orientation	= f.df((double) pos / divisions);
		accel		= f.ddf((double) pos / divisions);

		orientation.normalise();
		accel.normalise();

		// Find a vector in plane normal to orientation

		if (pos==-1)
		{
			if (orientation.x()>orientation.y() && orientation.x()>orientation.z())
				r = xProduct(orientation,VectorY);
			else
			if (orientation.y()>orientation.z())
				r = xProduct(orientation,VectorZ);
			else
				r = xProduct(orientation,VectorX);

			r.normalise();
		}
		else
		{
			double dot = orientation*r;
			double alpha = acos(dot) - Pi/2.0;

			Vector axis  = xProduct(orientation,r);
			axis.normalise();
			r = matrixRotate(axis,alpha*180.0/Pi)*r;
			r.normalise();
		}

		// Calculate vertices around circle

		if (pos>=0)
			for (vert=0;vert<samples;vert++)
			{
				Matrix rotate = matrixRotate(orientation,(double) vert * (360.0/samples));
				vertex[vert] = (rotate*r)*radius+position;
			}

		// Start face

		if (pos==1 || pos>=1 && convex)
		{
			Uint vBase = mesh->point().size();

			for (vert=0;vert<samples;vert++)
				mesh->point().push_back(Point(prev_vertex[vert],-prev_orientation));

			Uint vCenter = mesh->point().size();
			mesh->point().push_back(Point(prev_position,-prev_orientation));

			for (vert=0;vert<samples;vert++)
				mesh->triangle().push_back(
					Triangle(
						&mesh->point()[vert+vBase],
						&mesh->point()[((vert+1)%samples)+vBase],
						&mesh->point()[vCenter]
					)
				);
		}
		
		//
		// Connecting segments
		//

		if (pos>=0 && convex || pos==1 && !convex)
			for (vert=0;vert<samples;vert++)
				mesh->point().push_back(Point(prev_vertex[vert],prev_vertex[vert]-prev_position));
		
		if (pos>=0)
			for (vert=0;vert<samples;vert++)
				mesh->point().push_back(Point(vertex[vert],vertex[vert]-position));

		if (pos>0)
		{
			Uint vBase = mesh->point().size()-samples;
			Uint pvBase = vBase-samples;

			for (vert=0;vert<samples;vert++)
			{
				mesh->triangle().push_back(
					Triangle(
						&mesh->point()[vert+vBase],
						&mesh->point()[((vert+1)%samples)+vBase],
						&mesh->point()[vert+pvBase]
					)
				);

				mesh->triangle().push_back(
					Triangle(
						&mesh->point()[((vert+1)%samples)+vBase],
						&mesh->point()[((vert+1)%samples)+pvBase],
						&mesh->point()[vert+pvBase]
					)
				);
			}

//			#define SWEEP_POV_OUTPUT
			#ifdef SWEEP_POV_OUTPUT

			Vector core = position - prev_position;
			core.normalise();

			double angle = acos(core*VectorX)/Degr;
			Vector axis  = xProduct(core,VectorX);

			Matrix trans = matrixTranslate(prev_position);
			trans = matrixRotate(axis,angle) * trans;
			Matrix inv = trans.inverse();

			double maxx = (inv * prev_vertex[0]).x();
			double minx = (inv * vertex[0]).x();
	
			for (Uint v=0;v<samples;v++)
			{
				double x = (inv * prev_vertex[v]).x();

				maxx = MIN(x,maxx);

				x = (inv * vertex[v]).x();

				minx = MAX(x,minx);
			}

//			cout << minx << ' ' << maxx << endl;

			cout << "cylinder { " << endl;
			(trans * ( VectorX * minx)).writePOV(cout) << ',';
			(trans * ( VectorX * maxx)).writePOV(cout) << ',';
			cout << radius << '}' << endl;
			#endif

		}

		//
		// End face
		//

		if (pos==divisions || pos>=1 && convex)
		{
			Uint vBase = mesh->point().size();

			for (vert=0;vert<samples;vert++)
				mesh->point().push_back(Point(vertex[vert],orientation));

			Uint vCenter = mesh->point().size();
			mesh->point().push_back(Point(position,orientation));

			for (vert=0;vert<samples;vert++)
				mesh->triangle().push_back(
					Triangle(
						&mesh->point()[((vert+1)%samples)+vBase],
						&mesh->point()[vert+vBase],
						&mesh->point()[vCenter]
					)
				);
		}

		if (pos>0 && convex)
		{
			meshes.push_back(mesh);
			mesh = new Mesh();
		}

		prev_position = position;
		prev_orientation = orientation;

		Vector *tmp = prev_vertex;
		prev_vertex = vertex;
		vertex = tmp;
	}


	delete [] vertex;
	delete [] prev_vertex;

	if (!convex)
		meshes.push_back(mesh);
	else
		delete mesh;

	return meshes;
}



Mesh *MeshSweep(Func &f,const double radius,const int divisions,const int samples)
{
	vector<Mesh *> meshes = _MeshSweep(f,radius,divisions,samples);
	assert(meshes.size()==1);
	return meshes.front();
}
	
vector<Mesh *> 
MeshSweepConvex(Func &f,const double radius,const int divisions,const int samples)
{
	return _MeshSweep(f,radius,divisions,samples,true);
}




