#include <vector3.h>

#include <math.h>
#include <assert.h>

Vector::Vector()
{
	_vector[0] = _vector[1] = _vector[2] = 0.0;
}

Vector::Vector(double a, double b, double c)
{
	_vector[0] = a;
	_vector[1] = b;
	_vector[2] = c;
}

const double &
Vector::operator[](int n) const
{
	return _vector[n];
}

double &
Vector::operator[](int n)
{
	return _vector[n];
}

      double &Vector::x()       { return _vector[0]; }
const double &Vector::x() const { return _vector[0]; }
      double &Vector::y()       { return _vector[1]; }
const double &Vector::y() const { return _vector[1]; }
      double &Vector::z()       { return _vector[2]; }
const double &Vector::z() const { return _vector[2]; }

bool
Vector::operator==(const Vector &v)
{
	return 
		_vector[0] == v[0] &&
		_vector[1] == v[1] &&
		_vector[2] == v[2];
}

bool
Vector::operator==(const double &x)
{
	return 
		_vector[0] == x &&
		_vector[1] == x &&
		_vector[2] == x;
}

ostream &
operator<<(ostream &os, const Vector &x)
{
	os << '[';
	os << x[0] << ',';
	os << x[1] << ',';
	os << x[2];
	os << ']';

	return os;
}

ostream &
Vector::writePOV(ostream &os) const
{
	os << "< ";
	os << x() << ',';
	os << y() << ',';
	os << z() << " >";

	return os;
}

Vector
operator*(const double x, const Vector &v)
{
	return v*x;
}

Vector
operator*(const Vector &v, const double x)
{
	return Vector( x*v._vector[0], x*v._vector[1], x*v._vector[2]);
}

Vector
operator/(const Vector &v, const double x)
{
	assert(x != 0.0);

	double inv = 1.0/x;
	return v*inv;
}

Vector
Vector::operator-() const
{
	return Vector(-_vector[0], -_vector[1], -_vector[2]);
}

Vector
operator+(const Vector &v1, const Vector &v2)
{
	return Vector(v1[0] + v2[0], v1[1] + v2[1], v1[2] + v2[2]);
}

Vector
operator-(const Vector &v1, const Vector &v2)
{
	return Vector(v1[0] - v2[0], v1[1] - v2[1], v1[2] - v2[2]);
}

void
Vector::normalise()
{
	double mag = sqrt(norm());

	if (mag == 0.0)
		return;

	double magInv = 1.0/mag;

	_vector[0] *= magInv;
	_vector[1] *= magInv;
	_vector[2] *= magInv;
}

double
Vector::norm() const
{
	return ((*this)*(*this));
}

Vector &
Vector::operator=(const Vector &x)
{
	_vector[0] = x[0];
	_vector[1] = x[1];
	_vector[2] = x[2];

	return *this;
}

double
Vector::operator*(const Vector &x) const
{
	return _vector[0]*x[0] + _vector[1]*x[1] + _vector[2]*x[2];
}

Vector & 
Vector::operator+=(const Vector &v)
{
	_vector[0] += v[0];
	_vector[1] += v[1];
	_vector[2] += v[2];

	return *this;
}

Vector & 
Vector::operator-=(const Vector &x)
{
	_vector[0] -= x[0];
	_vector[1] -= x[1];
	_vector[2] -= x[2];

	return *this;
}

double
Vector::dist(const Vector &x) const
{
	return ((*this)-x).norm();
}

Vector 
xProduct(const Vector &v1, const Vector &v2)
{
	return Vector(v1[1]*v2[2] - v1[2]*v2[1],
				  v1[2]*v2[0] - v1[0]*v2[2],
				  v1[0]*v2[1] - v1[1]*v2[0]);
}

/////////////////////////
//
// OpenGL

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>

void
Vector::glVertex() const
{
	glVertex3dv(_vector);
}

void
Vector::glNormal() const
{
	glNormal3dv(_vector);
}

void
Vector::glColor() const
{
	glColor3dv(_vector);
}

Vector::operator double *()
{
	return (double *) _vector;
}

